# 11 "tools/coqdoc/cpretty.mll"
 
  open Printf
  open Lexing

  (* A function that emulates Lexing.new_line (which does not exist in OCaml < 3.11.0) *)
  let new_line lexbuf =
    let pos = lexbuf.lex_curr_p in
      lexbuf.lex_curr_p <- { pos with
	pos_lnum = pos.pos_lnum + 1;
	pos_bol = pos.pos_cnum }

  (* A list function we need *)
  let rec take n ls =
    if n = 0 then [] else
      match ls with
        | [] -> []
        | (l :: ls) -> l :: (take (n-1) ls)

  (* count the number of spaces at the beginning of a string *)
  let count_spaces s =
    let n = String.length s in
    let rec count c i =
      if i == n then c,i else match s.[i] with
	| '\t' -> count (c + (8 - (c mod 8))) (i + 1)
	| ' ' -> count (c + 1) (i + 1)
	| _ -> c,i
    in
      count 0 0

  let remove_newline s =
    let n = String.length s in
    let rec count i = if i == n || s.[i] <> '\n' then i else count (i + 1) in
    let i = count 0 in
    i, String.sub s i (n - i)

  let count_dashes s =
    let c = ref 0 in
    for i = 0 to String.length s - 1 do if s.[i] = '-' then incr c done;
    !c

  let cut_head_tail_spaces s =
    let n = String.length s in
    let rec look_up i = if i == n || s.[i] <> ' ' then i else look_up (i+1) in
    let rec look_dn i = if i == -1 || s.[i] <> ' ' then i else look_dn (i-1) in
    let l = look_up 0 in
    let r = look_dn (n-1) in
    if l <= r then String.sub s l (r-l+1) else s

  let sec_title s =
    let rec count lev i =
      if s.[i] = '*' then
	count (succ lev) (succ i)
      else
	let t = String.sub s i (String.length s - i) in
	lev, cut_head_tail_spaces t
    in
    count 0 (String.index s '*')

  let strip_eol s =
    let eol = s.[String.length s - 1] = '\n' in
    (eol, if eol then String.sub s 1 (String.length s - 1) else s)


  let formatted = ref false
  let brackets = ref 0
  let comment_level = ref 0
  let in_proof = ref None
  let in_emph = ref false

  let in_env start stop = 
    let r = ref false in
    let start_env () = r := true; start () in
    let stop_env () = if !r then stop (); r := false in
      (fun x -> !r), start_env, stop_env
      
  let in_emph, start_emph, stop_emph = in_env Output.start_emph Output.stop_emph
  let in_quote, start_quote, stop_quote = in_env Output.start_quote Output.stop_quote

  let url_buffer = Buffer.create 40
  let url_name_buffer = Buffer.create 40

  let backtrack lexbuf = lexbuf.lex_curr_pos <- lexbuf.lex_start_pos;
    lexbuf.lex_curr_p <- lexbuf.lex_start_p

  let backtrack_past_newline lexbuf =
    let buf = lexeme lexbuf in
    let splits = Str.bounded_split_delim (Str.regexp "['\n']") buf 2 in
      match splits with
        | [] -> ()
        | (_ :: []) -> ()
        | (s1 :: rest :: _) ->
            let length_skip = 1 + String.length s1 in
              lexbuf.lex_curr_pos <- lexbuf.lex_start_pos + length_skip

  let is_space = function ' ' | '\t' | '\n' | '\r' -> true | _ -> false

  (* saving/restoring the PP state *)

  type state = {
    st_gallina : bool;
    st_light : bool
  }

  let state_stack = Stack.create ()

  let save_state () =
    Stack.push { st_gallina = !Cdglobals.gallina; st_light = !Cdglobals.light } state_stack

  let restore_state () =
    let s = Stack.pop state_stack in
    Cdglobals.gallina := s.st_gallina;
    Cdglobals.light := s.st_light

  let without_ref r f x = save_state (); r := false; f x; restore_state ()

  let without_gallina = without_ref Cdglobals.gallina

  let without_light = without_ref Cdglobals.light

  let show_all f = without_gallina (without_light f)

  let begin_show () = save_state (); Cdglobals.gallina := false; Cdglobals.light := false
  let end_show () = restore_state ()

  (* Reset the globals *)

  let reset () =
    formatted := false;
    brackets := 0;
    comment_level := 0

  (* erasing of Section/End *)

  let section_re = Str.regexp "[ \t]*Section"
  let end_re = Str.regexp "[ \t]*End"
  let is_section s = Str.string_match section_re s 0
  let is_end s = Str.string_match end_re s 0

  let sections_to_close = ref 0

  let section_or_end s =
    if is_section s then begin
      incr sections_to_close; true
    end else if is_end s then begin
      if !sections_to_close > 0 then begin
	decr sections_to_close; true
      end else
	false
    end else
      true

  (* for item lists *)
  type list_compare =
      | Before
      | StartLevel of int
      | InLevel of int * bool

  (* Before : we're before any levels
     StartLevel : at the same column as the dash in a level
     InLevel : after the dash of this level, but before any deeper dashes.
               bool is true if this is the last level *)
  let find_level levels cur_indent =
    match levels with
    | [] -> Before
    | (l::ls) ->
        if cur_indent < l then Before
        else
          (* cur_indent will never be less than the head of the list *)
          let rec findind ls n =
            match ls with
            | [] -> InLevel (n,true)
            | (l :: []) -> if cur_indent = l then StartLevel n
                           else InLevel (n,true)
            | (l1 :: l2 :: ls) ->
                if cur_indent = l1 then StartLevel n
                else if cur_indent < l2 then InLevel (n,false)
                     else findind (l2 :: ls) (n+1)
          in
            findind (l::ls) 1

  type is_start_list =
    | Rule
    | List of int
    | Neither

  let check_start_list str =
    let n_dashes = count_dashes str in
    let (n_spaces,_) = count_spaces str in
      if n_dashes >= 4 && not !Cdglobals.plain_comments then
        Rule
      else
        if n_dashes = 1 && not !Cdglobals.plain_comments then
          List n_spaces
        else
          Neither

   (* examine a string for subtitleness *)
  let subtitle m s =
    match Str.split_delim (Str.regexp ":") s with
    | [] -> false
    | (name::_) ->
        if (cut_head_tail_spaces name) = m then
          true
        else
          false


  (* tokens pretty-print *)

  let token_buffer = Buffer.create 1024

  let token_re =
    Str.regexp "[ \t]*(\\*\\*[ \t]+printing[ \t]+\\([^ \t]+\\)"
  let printing_token_re =
    Str.regexp
      "[ \t]*\\(\\(%\\([^%]*\\)%\\)\\|\\(\\$[^$]*\\$\\)\\)?[ \t]*\\(#\\(\\(&#\\|[^#]\\)*\\)#\\)?"

  let add_printing_token toks pps =
    try
      if Str.string_match token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	if Str.string_match printing_token_re pps 0 then
	  let pp =
	    (try Some (Str.matched_group 3 pps) with _ ->
             try Some (Str.matched_group 4 pps) with _ -> None),
	    (try Some (Str.matched_group 6 pps) with _ -> None)
	  in
	  Output.add_printing_token tok pp
    with _ ->
      ()

  let remove_token_re =
    Str.regexp
      "[ \t]*(\\*\\*[ \t]+remove[ \t]+printing[ \t]+\\([^ \t]+\\)[ \t]*\\*)"

  let remove_printing_token toks =
    try
      if Str.string_match remove_token_re toks 0 then
	let tok = Str.matched_group 1 toks in
	Output.remove_printing_token tok
    with _ ->
      ()

  let extract_ident_re = Str.regexp "([ \t]*\\([^ \t]+\\)[ \t]*:="
  let extract_ident s =
    assert (String.length s >= 3);
    if Str.string_match extract_ident_re s 0 then
      Str.matched_group 1 s
    else
      String.sub s 1 (String.length s - 3)

  let output_indented_keyword s lexbuf =
    let nbsp,isp = count_spaces s in
    Output.indentation nbsp;
    let s = String.sub s isp (String.length s - isp) in
    Output.keyword s (lexeme_start lexbuf + isp)


# 261 "tools/coqdoc/cpretty.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\235\255\236\255\000\000\000\000\001\000\000\000\000\000\
    \002\000\002\000\005\000\004\000\007\000\022\000\014\000\027\000\
    \024\000\017\000\004\000\031\000\036\000\037\000\052\000\000\000\
    \001\000\002\000\141\000\010\000\021\000\022\000\057\000\041\000\
    \058\000\064\000\065\000\024\000\075\000\083\000\065\000\073\000\
    \063\000\080\000\008\000\128\000\056\000\128\000\143\000\001\000\
    \005\000\221\000\223\000\036\000\064\000\227\000\007\000\254\255\
    \226\000\074\000\074\000\232\000\084\000\084\000\079\000\091\000\
    \089\000\098\000\009\000\012\000\013\000\014\000\015\000\093\000\
    \092\000\087\000\114\000\016\000\239\000\102\000\128\000\132\000\
    \129\000\141\000\139\000\147\000\017\000\018\000\019\000\021\000\
    \022\000\253\000\011\000\239\255\149\000\152\000\150\000\024\000\
    \008\001\159\000\161\000\167\000\171\000\043\001\014\000\025\000\
    \252\255\009\000\165\000\159\000\044\001\017\000\045\001\251\255\
    \019\000\179\000\027\000\047\001\176\000\170\000\163\000\051\001\
    \033\000\057\001\250\255\027\000\216\000\244\255\173\000\218\000\
    \001\001\170\000\231\000\208\000\214\000\209\000\225\000\235\000\
    \233\000\232\000\235\000\087\001\253\255\246\000\247\000\004\001\
    \245\000\239\000\247\255\241\000\250\000\252\000\003\001\255\000\
    \253\000\014\001\255\000\018\001\002\001\015\001\003\001\021\001\
    \023\001\024\001\027\001\011\001\011\001\015\001\029\001\011\001\
    \016\001\034\001\029\001\024\001\026\001\022\001\023\001\038\001\
    \248\255\025\001\030\001\032\001\038\001\045\001\048\001\038\001\
    \035\001\049\001\042\001\034\001\050\001\038\001\044\001\057\001\
    \058\001\056\001\061\001\051\001\060\001\249\255\064\001\055\001\
    \068\001\069\001\057\001\071\001\073\001\063\001\163\001\164\001\
    \064\001\060\001\060\001\080\001\068\001\080\001\079\001\073\001\
    \081\001\084\001\073\001\079\001\074\001\084\001\092\001\087\001\
    \077\001\185\001\189\001\083\001\097\001\084\001\093\001\102\001\
    \094\001\089\001\091\001\091\001\103\001\101\001\098\001\094\001\
    \113\001\097\001\110\001\098\001\117\001\118\001\123\001\112\001\
    \124\001\123\001\126\001\110\001\110\001\128\001\128\001\223\001\
    \224\001\109\001\119\001\133\001\115\001\242\255\126\001\122\001\
    \134\001\124\001\120\001\148\001\142\001\131\001\148\001\127\001\
    \141\001\135\001\149\001\133\001\138\001\153\001\161\001\255\001\
    \000\002\155\001\170\001\168\001\157\001\173\001\155\001\168\001\
    \159\001\160\001\175\001\161\001\172\001\160\001\178\001\015\002\
    \016\002\162\001\167\001\170\001\188\001\187\001\173\001\185\001\
    \180\001\182\001\179\001\197\001\182\001\181\001\180\001\198\001\
    \200\001\036\002\037\002\194\001\190\001\210\001\193\001\205\001\
    \200\001\202\001\216\001\214\001\198\001\219\001\203\001\209\001\
    \208\001\206\001\205\001\210\001\215\001\225\001\230\001\229\001\
    \213\001\225\001\222\001\224\001\239\001\221\001\233\001\228\001\
    \230\001\227\001\235\001\232\001\235\001\230\001\082\002\084\002\
    \237\001\240\001\239\001\238\001\244\001\253\001\235\001\240\001\
    \000\002\244\001\249\001\245\001\250\001\248\001\002\002\249\001\
    \009\002\003\002\250\001\012\002\008\002\017\002\019\002\015\002\
    \024\002\006\002\022\002\027\002\015\002\011\002\015\002\031\002\
    \015\002\029\002\021\002\016\002\036\002\021\002\018\002\035\002\
    \035\002\027\002\031\002\040\002\031\002\027\002\027\002\030\002\
    \028\002\047\002\031\002\037\002\035\002\035\002\041\002\038\002\
    \051\002\055\002\043\002\059\002\044\002\046\002\054\002\070\002\
    \052\002\060\002\055\002\072\002\069\002\072\002\079\002\061\002\
    \073\002\068\002\070\002\066\002\243\255\071\002\085\002\080\002\
    \069\002\080\002\084\002\081\002\086\002\076\002\097\002\081\002\
    \095\002\188\002\189\002\089\002\095\002\100\002\086\002\092\002\
    \104\002\088\002\094\002\103\002\103\002\094\002\101\002\103\002\
    \111\002\112\002\102\002\113\002\100\002\208\002\210\002\106\002\
    \122\002\112\002\122\002\122\002\111\002\113\002\113\002\133\002\
    \121\002\133\002\132\002\131\002\226\002\227\002\135\002\136\002\
    \129\002\142\002\125\002\144\002\245\255\141\002\137\002\139\002\
    \138\002\139\002\153\002\152\002\157\002\145\002\141\002\145\002\
    \163\002\147\002\162\002\154\002\151\002\157\002\155\002\169\002\
    \164\002\158\002\170\002\177\002\012\003\013\003\168\002\164\002\
    \184\002\166\002\178\002\173\002\175\002\185\002\179\002\184\002\
    \175\002\192\002\173\002\195\002\181\002\180\002\032\003\033\003\
    \190\002\188\002\198\002\194\002\189\002\202\002\198\002\206\002\
    \045\003\047\003\205\002\204\002\204\002\215\002\203\002\219\002\
    \214\002\211\002\213\002\209\002\220\002\208\002\226\002\214\002\
    \214\002\229\002\216\002\224\002\222\002\228\002\232\002\239\002\
    \234\002\224\002\076\003\077\003\229\002\241\002\228\002\237\002\
    \247\002\239\002\234\002\236\002\247\002\242\002\244\002\000\003\
    \252\002\010\000\000\003\004\003\248\002\243\002\245\002\244\002\
    \252\002\006\003\255\002\015\003\016\003\020\000\010\003\022\003\
    \005\003\006\003\005\003\024\003\008\003\008\003\012\003\026\003\
    \027\003\020\003\029\003\013\003\032\003\019\003\034\003\020\003\
    \027\003\018\003\040\003\024\003\029\003\032\003\039\003\035\003\
    \030\003\048\003\032\003\044\003\039\003\041\003\052\003\039\003\
    \049\003\045\003\053\003\043\003\057\003\057\003\060\003\048\003\
    \048\003\061\003\051\003\063\003\059\003\062\003\057\003\053\003\
    \166\003\246\255\167\003\062\003\073\003\063\003\076\003\076\003\
    \072\003\080\003\085\003\069\003\085\003\087\003\012\000\077\003\
    \078\003\078\003\094\003\083\003\184\003\185\003\078\003\099\003\
    \085\003\089\003\103\003\081\003\105\003\093\003\108\003\108\003\
    \105\003\100\003\113\003\133\003\119\003\142\003\134\003\166\003\
    \110\003\130\003\134\003\133\003\125\003\124\003\139\003\133\003\
    \133\003\131\003\136\003\152\003\146\003\171\003\163\003\158\003\
    \159\003\171\003\178\003\013\004\014\004\170\003\166\003\191\003\
    \173\003\185\003\180\003\182\003\192\003\197\003\188\003\186\003\
    \201\003\202\003\200\003\187\003\204\003\196\003\187\003\209\003\
    \193\003\198\003\200\003\207\003\203\003\198\003\216\003\200\003\
    \212\003\207\003\209\003\221\003\221\003\210\003\227\003\205\003\
    \211\003\219\003\227\003\064\004\066\004\221\003\236\003\234\003\
    \223\003\242\003\224\003\237\003\228\003\229\003\244\003\230\003\
    \241\003\229\003\248\003\085\004\086\004\233\003\239\003\242\003\
    \004\004\003\004\243\003\255\003\250\003\252\003\247\003\011\004\
    \251\003\250\003\014\004\002\004\016\004\046\004\000\004\014\004\
    \011\004\017\004\012\004\016\004\017\004\029\004\013\004\007\004\
    \009\004\019\004\021\004\028\004\024\004\019\004\036\004\020\004\
    \039\004\023\004\035\004\023\004\042\004\043\004\044\004\047\004\
    \032\004\032\004\036\004\050\004\032\004\037\004\055\004\052\004\
    \047\004\049\004\046\004\047\004\052\004\054\004\060\004\069\004\
    \072\004\062\004\059\004\073\004\066\004\058\004\074\004\062\004\
    \077\004\064\004\074\004\070\004\070\004\080\004\082\004\070\004\
    \072\004\083\004\073\004\085\004\080\004\082\004\078\004\074\004\
    \097\004\081\004\097\004\099\004\013\000\089\004\090\004\090\004\
    \087\004\107\004\096\004\105\004\091\004\107\004\095\004\099\004\
    \114\004\104\004\115\004\117\004\107\004\207\004\208\004\108\004\
    \104\004\104\004\124\004\112\004\124\004\123\004\114\004\127\004\
    \119\004\123\004\120\004\126\004\116\004\128\004\123\004\125\004\
    \139\004\123\004\137\004\232\004\233\004\132\004\134\004\130\004\
    \130\004\150\004\138\004\150\004\149\004\151\004\135\004\145\004\
    \153\004\143\004\159\004\153\004\145\004\162\004\023\000\153\004\
    \165\004\148\004\151\004\150\004\169\004\153\004\153\004\157\004\
    \171\004\172\004\165\004\174\004\162\004\161\004\173\004\168\004\
    \170\004\183\004\179\004\219\004\184\004\188\004\176\004\171\004\
    \173\004\171\004\177\004\187\004\177\004\193\004\198\004\196\004\
    \180\004\201\004\185\004\191\004\190\004\195\004\203\004\208\004\
    \208\004\192\004\204\004\200\004\202\004\216\004\198\004\210\004\
    \205\004\208\004\204\004\212\004\201\004\221\004\223\004\212\004\
    \224\004\227\004\220\004\217\004\213\004\232\004\216\004\229\004\
    \217\004\235\004\236\004\241\004\229\004\241\004\240\004\244\004\
    \228\004\228\004\246\004\246\004\083\005\084\005\229\004\239\004\
    \251\004\243\004\238\004\250\004\240\004\236\004\250\004\254\004\
    \005\005\000\005\246\004\251\004\253\004\249\004\006\005\010\005\
    \253\004\012\005\255\004\003\005\008\005\020\005\016\005\011\005\
    \014\005\009\005\008\005\010\005\027\005\011\005\021\005\029\005\
    \034\005\019\005\016\005\033\005\021\005\031\005\040\005\040\005\
    \031\005\039\005\026\005\037\005\027\005\028\005\048\005\037\005\
    \046\005\032\005\048\005\036\005\035\005\032\005\037\005\145\005\
    \146\005\058\005\049\005\053\005\056\005\063\005\045\005\057\005\
    \052\005\054\005\068\005\050\005\062\005\058\005\060\005\066\005\
    \072\005\057\005\069\005\064\005\076\005\074\005\064\005\065\005\
    \081\005\083\005\070\005\070\005\072\005\078\005\081\005\076\005\
    \072\005\092\005\094\005\095\005\079\005\089\005\097\005\085\005\
    \095\005\104\005\104\005\095\005\103\005\090\005\197\005\242\255\
    \199\005\112\005\100\005\101\005\105\005\117\005\119\005\119\005\
    \121\005\122\005\105\005\127\005\187\005\131\005\149\005\129\005\
    \137\005\141\005\120\005\184\005\187\005\251\255\049\000\023\006\
    \083\000\001\000\239\005\033\006\085\000\252\255\054\000\038\006\
    \254\255\129\005\210\005\135\005\144\005\136\005\151\005\173\005\
    \181\005\174\005\028\006\029\006\191\005\203\005\196\005\205\005\
    \214\005\208\005\203\005\205\005\250\255\206\005\218\005\216\005\
    \215\005\211\005\230\005\214\005\228\005\216\005\234\005\247\255\
    \235\005\240\005\228\005\240\005\239\005\242\005\226\005\226\005\
    \244\005\244\005\081\006\082\006\227\005\237\005\250\005\232\005\
    \246\255\243\005\238\005\250\005\240\005\236\005\249\005\006\006\
    \007\006\007\006\010\006\008\006\250\005\010\006\003\006\018\006\
    \018\006\018\006\007\006\024\006\003\006\017\006\009\006\024\006\
    \010\006\015\006\023\006\031\006\127\006\128\006\027\006\042\006\
    \040\006\029\006\045\006\027\006\040\006\031\006\032\006\047\006\
    \033\006\044\006\032\006\050\006\143\006\144\006\034\006\039\006\
    \042\006\060\006\059\006\045\006\057\006\052\006\054\006\051\006\
    \071\006\055\006\054\006\053\006\071\006\073\006\168\006\169\006\
    \063\006\063\006\083\006\065\006\077\006\072\006\074\006\088\006\
    \086\006\070\006\091\006\075\006\081\006\080\006\078\006\077\006\
    \082\006\087\006\095\006\101\006\100\006\086\006\100\006\095\006\
    \098\006\112\006\094\006\106\006\101\006\104\006\100\006\108\006\
    \105\006\108\006\103\006\211\006\212\006\111\006\113\006\112\006\
    \111\006\117\006\126\006\108\006\113\006\129\006\117\006\122\006\
    \118\006\123\006\121\006\131\006\122\006\138\006\132\006\123\006\
    \141\006\136\006\145\006\148\006\144\006\153\006\135\006\151\006\
    \156\006\144\006\140\006\144\006\160\006\144\006\158\006\150\006\
    \145\006\165\006\150\006\147\006\164\006\164\006\156\006\160\006\
    \169\006\160\006\156\006\156\006\159\006\157\006\176\006\160\006\
    \166\006\164\006\165\006\171\006\167\006\180\006\185\006\172\006\
    \187\006\174\006\175\006\185\006\198\006\185\006\190\006\184\006\
    \201\006\198\006\201\006\208\006\190\006\202\006\197\006\199\006\
    \195\006\245\255\214\006\202\006\243\006\217\006\202\006\203\006\
    \218\006\205\006\207\006\216\006\213\006\219\006\214\006\218\006\
    \219\006\232\006\216\006\210\006\213\006\222\006\224\006\231\006\
    \227\006\222\006\239\006\223\006\242\006\226\006\238\006\227\006\
    \245\006\246\006\247\006\250\006\234\006\234\006\238\006\253\006\
    \235\006\240\006\003\007\254\006\249\006\251\006\251\006\253\006\
    \003\007\010\007\013\007\003\007\000\007\014\007\007\007\255\006\
    \015\007\003\007\007\007\021\007\016\007\005\007\016\007\020\007\
    \017\007\022\007\012\007\033\007\017\007\031\007\124\007\125\007\
    \025\007\031\007\036\007\022\007\028\007\040\007\024\007\030\007\
    \039\007\039\007\030\007\037\007\039\007\047\007\048\007\038\007\
    \049\007\036\007\144\007\146\007\042\007\058\007\048\007\058\007\
    \058\007\047\007\049\007\049\007\069\007\057\007\069\007\068\007\
    \067\007\162\007\163\007\071\007\072\007\065\007\078\007\061\007\
    \080\007\248\255\077\007\073\007\075\007\074\007\075\007\089\007\
    \088\007\093\007\081\007\077\007\081\007\099\007\083\007\098\007\
    \090\007\087\007\093\007\091\007\105\007\100\007\094\007\106\007\
    \113\007\204\007\205\007\104\007\100\007\120\007\102\007\114\007\
    \109\007\111\007\121\007\115\007\120\007\111\007\128\007\109\007\
    \131\007\117\007\116\007\224\007\225\007\126\007\124\007\134\007\
    \130\007\125\007\138\007\134\007\142\007\237\007\239\007\141\007\
    \140\007\140\007\151\007\139\007\155\007\150\007\147\007\149\007\
    \145\007\156\007\144\007\162\007\150\007\150\007\165\007\152\007\
    \160\007\158\007\164\007\168\007\175\007\170\007\160\007\012\008\
    \013\008\165\007\177\007\164\007\173\007\183\007\175\007\170\007\
    \172\007\183\007\178\007\180\007\192\007\188\007\228\007\193\007\
    \197\007\185\007\180\007\183\007\183\007\189\007\202\007\192\007\
    \208\007\210\007\245\007\205\007\217\007\200\007\201\007\200\007\
    \219\007\203\007\203\007\207\007\221\007\222\007\216\007\225\007\
    \209\007\227\007\214\007\229\007\215\007\216\007\213\007\219\007\
    \071\008\072\008\240\007\231\007\235\007\238\007\245\007\227\007\
    \239\007\234\007\236\007\249\255\250\007\232\007\244\007\239\007\
    \241\007\252\007\239\007\250\007\244\007\253\007\243\007\001\008\
    \005\008\005\008\101\008\102\008\253\007\009\008\255\007\013\008\
    \013\008\009\008\017\008\007\008\007\008\018\008\008\008\020\008\
    \015\008\017\008\012\008\008\008\031\008\015\008\031\008\035\008\
    \060\008\026\008\027\008\027\008\043\008\032\008\133\008\157\008\
    \026\008\047\008\033\008\035\008\052\008\028\008\053\008\043\008\
    \051\008\057\008\054\008\049\008\062\008\080\008\066\008\073\008\
    \074\008\082\008\070\008\084\008\090\008\091\008\082\008\080\008\
    \094\008\088\008\088\008\084\008\088\008\104\008\088\008\103\008\
    \095\008\090\008\091\008\103\008\110\008\203\008\222\008\099\008\
    \095\008\116\008\098\008\110\008\105\008\107\008\131\008\145\008\
    \136\008\137\008\150\008\151\008\149\008\134\008\150\008\143\008\
    \135\008\157\008\141\008\146\008\148\008\155\008\151\008\146\008\
    \164\008\148\008\160\008\155\008\157\008\169\008\169\008\158\008\
    \175\008\153\008\159\008\167\008\175\008\014\009\015\009\166\008\
    \184\008\182\008\171\008\190\008\172\008\185\008\176\008\177\008\
    \192\008\178\008\189\008\177\008\195\008\032\009\034\009\178\008\
    \185\008\190\008\208\008\207\008\191\008\203\008\198\008\200\008\
    \195\008\215\008\199\008\198\008\218\008\206\008\220\008\251\008\
    \204\008\217\008\215\008\221\008\216\008\220\008\221\008\233\008\
    \217\008\211\008\214\008\223\008\225\008\232\008\228\008\223\008\
    \240\008\224\008\243\008\227\008\239\008\227\008\245\008\247\008\
    \248\008\251\008\235\008\236\008\240\008\254\008\236\008\244\008\
    \007\009\002\009\253\008\000\009\252\008\253\008\002\009\004\009\
    \010\009\017\009\020\009\010\009\007\009\021\009\014\009\006\009\
    \022\009\010\009\025\009\012\009\022\009\018\009\018\009\028\009\
    \030\009\018\009\020\009\031\009\021\009\033\009\028\009\030\009\
    \026\009\022\009\045\009\029\009\045\009\047\009\072\009\038\009\
    \039\009\039\009\036\009\056\009\045\009\054\009\040\009\056\009\
    \044\009\048\009\063\009\053\009\064\009\066\009\056\009\156\009\
    \157\009\057\009\053\009\053\009\073\009\061\009\073\009\072\009\
    \063\009\076\009\068\009\072\009\069\009\075\009\065\009\077\009\
    \072\009\074\009\088\009\072\009\086\009\181\009\182\009\081\009\
    \083\009\079\009\079\009\099\009\087\009\099\009\098\009\100\009\
    \084\009\094\009\102\009\092\009\108\009\102\009\094\009\111\009\
    \143\009\103\009\115\009\100\009\101\009\100\009\119\009\103\009\
    \103\009\107\009\121\009\122\009\115\009\124\009\112\009\111\009\
    \123\009\118\009\121\009\133\009\129\009\169\009\134\009\138\009\
    \126\009\121\009\123\009\121\009\127\009\137\009\127\009\143\009\
    \148\009\146\009\130\009\151\009\135\009\141\009\140\009\145\009\
    \153\009\159\009\158\009\142\009\155\009\150\009\152\009\166\009\
    \148\009\160\009\155\009\158\009\154\009\162\009\151\009\171\009\
    \173\009\162\009\174\009\177\009\170\009\167\009\163\009\182\009\
    \166\009\179\009\167\009\185\009\186\009\191\009\179\009\191\009\
    \190\009\194\009\178\009\178\009\196\009\196\009\033\010\034\010\
    \179\009\189\009\201\009\193\009\188\009\200\009\190\009\186\009\
    \200\009\204\009\211\009\206\009\196\009\201\009\203\009\199\009\
    \212\009\216\009\203\009\218\009\205\009\209\009\214\009\226\009\
    \222\009\217\009\220\009\215\009\214\009\216\009\233\009\217\009\
    \227\009\235\009\240\009\225\009\222\009\239\009\227\009\237\009\
    \246\009\246\009\237\009\245\009\232\009\243\009\233\009\234\009\
    \254\009\243\009\252\009\238\009\254\009\242\009\252\009\002\010\
    \243\009\255\009\250\009\006\010\002\010\248\009\249\009\009\010\
    \011\010\255\009\252\009\254\009\006\010\010\010\005\010\011\010\
    \002\010\024\010\008\010\013\010\015\010\022\010\018\010\013\010\
    \031\010\015\010\027\010\022\010\024\010\024\010\039\010\029\010\
    \040\010\042\010\032\010\132\010\133\010\033\010\029\010\029\010\
    \049\010\037\010\049\010\048\010\030\010\050\010\052\010\053\010\
    \037\010\047\010\055\010\043\010\053\010\062\010\062\010\053\010\
    \061\010\048\010\157\010\249\255\250\255\251\255\108\010\124\010\
    \162\010\164\010\169\010\171\010\173\010\178\010\182\010\212\010\
    \191\010\199\010\203\010\187\010\183\010\255\255\227\010\206\010\
    \233\010\249\255\217\010\166\010\088\010\125\010\255\255\238\010\
    \221\010\245\010\247\010\253\255\178\010\252\255\180\010\236\010\
    \250\255\251\255\241\010\021\011\233\255\234\255\214\011\103\010\
    \109\010\172\010\184\010\197\010\203\010\243\010\214\010\166\010\
    \173\010\168\010\022\011\042\012\255\255\239\255\252\255\054\011\
    \005\011\254\255\003\011\253\255\009\011\251\255\249\255\011\011\
    \253\010\240\010\137\011\139\011\146\011\152\011\249\010\248\255\
    \244\255\242\255\240\255\142\011\237\255\004\011\253\255\254\255\
    \255\255\017\011\253\255\254\255\255\255\138\011\251\255\252\255\
    \006\011\009\011\253\255\254\255\172\011\251\255\252\255\239\010\
    \240\010\037\011\242\010\243\010\157\011\255\255\040\011\181\011\
    \253\255\193\010\253\255\183\010\255\255\196\010\254\255\255\255\
    \156\012\248\255\188\011\070\011\154\012\220\012\029\013\130\010\
    \182\013\251\255\025\011\037\011\254\255\255\255\253\255\252\255\
    \240\013\075\011\066\014\113\014\177\014\154\010\151\012\241\014\
    \045\011\114\011\086\011\032\015\079\015\143\015\155\010\006\014\
    \207\015\253\011\071\012\134\016\251\255\180\012\134\011\218\011\
    \139\016\144\016\254\255\220\011\252\255\146\016\241\255\242\255\
    \103\011\149\011\244\255\080\011\116\011\118\011\069\011\122\011\
    \122\011\255\255\073\012\254\255\162\011\184\012\196\011\251\255\
    \165\011\249\255\247\255\245\255\215\011\252\255\146\011\254\255\
    \213\013\255\255\185\011\098\014\253\255\137\016\254\255\205\011\
    \180\016\241\255\242\255\243\255\211\011\158\016\018\017\082\017\
    \025\011\235\017\142\018\169\011\152\016\252\255\142\011\010\018\
    \229\011\147\011\159\012\096\018\246\011\118\014\254\255\247\011\
    \144\014\253\255\248\011\244\255\251\255\166\011\006\012\104\017\
    \181\011\149\014\250\255\009\012\234\011\151\014\248\255\200\018\
    \227\011\026\019\222\017\090\019\121\011\243\019\150\020\057\021\
    \220\021\064\014\238\019\023\016\224\012\111\013\128\020\035\021\
    \215\021\123\011\127\022\236\013\105\022\191\022\255\022\125\011\
    \222\016\063\023\238\017\037\014\222\018\127\023\147\018\099\016\
    \241\018\191\023\232\018\248\016\168\018\253\255\254\255\255\255\
    \208\013\253\255\055\012\255\255\241\014\253\255\068\012\255\255\
    \184\013\254\255\094\012\255\255\142\016\057\012\070\012\177\012\
    \129\017\066\012\085\012\086\012\155\016\151\012\206\016\185\012\
    \185\013\254\255\255\255\169\012\080\013\201\012\055\021\253\255\
    \120\012\255\255\207\012\208\016\209\012\169\018\155\012\007\014\
    \254\255\240\012\118\024\254\255\104\014\147\016\057\021\220\021\
    \119\024\255\255\157\016\162\016\124\024\102\017\056\021\255\255\
    \027\019\126\024\078\019\231\020\254\255\128\024\219\021\150\022\
    \189\013\254\255\253\012\254\012\101\017\101\019\154\014\123\017\
    \125\024\254\255\242\013\103\024\240\024\048\025\112\012\222\025\
    \126\017\121\024\024\026\188\026\249\013\166\026\252\026\060\027\
    \162\012\199\020\124\027\058\021\169\022\062\016\171\027\218\027\
    \026\028\163\012\200\028\164\017\178\028\008\029\072\029\179\012\
    \167\024\136\029\131\022\195\022\205\024\200\029\245\023\052\024\
    \224\024\008\030\064\025\193\024\140\016\255\255\148\016\153\016\
    ";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\
    \000\000\020\000\020\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\018\000\255\255\255\255\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\015\000\255\255\014\000\014\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\017\000\255\255\255\255\
    \255\255\016\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\006\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\011\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\012\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\011\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\012\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\012\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\012\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\012\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\012\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\011\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\011\000\255\255\255\255\
    \011\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\013\000\013\000\013\000\255\255\013\000\000\000\
    \013\000\255\255\255\255\255\255\255\255\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \010\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\007\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\010\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\010\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\010\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\008\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\008\000\255\255\255\255\255\255\255\255\006\000\001\000\
    \002\000\006\000\006\000\006\000\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\003\000\
    \255\255\255\255\006\000\006\000\006\000\006\000\255\255\006\000\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\255\255\022\000\022\000\
    \020\000\022\000\014\000\012\000\010\000\022\000\022\000\022\000\
    \005\000\000\000\022\000\017\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\009\000\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\019\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \004\000\000\000\255\255\255\255\255\255\255\255\255\255\004\000\
    \004\000\004\000\255\255\255\255\001\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\001\000\255\255\255\255\255\255\255\255\
    \006\000\255\255\006\000\007\000\007\000\007\000\007\000\007\000\
    \005\000\255\255\007\000\007\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\005\000\002\000\005\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \014\000\012\000\255\255\009\000\007\000\005\000\003\000\014\000\
    \014\000\255\255\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\003\000\255\255\
    \003\000\255\255\255\255\001\000\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\014\000\014\000\014\000\014\000\
    \014\000\010\000\010\000\014\000\014\000\255\255\014\000\000\000\
    \014\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\004\000\255\255\006\000\
    \255\255\255\255\255\255\255\255\008\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\010\000\010\000\010\000\
    \009\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\010\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\001\000\255\255\255\255\255\255\001\000\255\255\
    \255\255\255\255\001\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\001\000\000\000\255\255\255\255\255\255\
    \002\000\255\255\002\000\002\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\001\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\001\000\001\000\001\000\001\000\001\000\001\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\001\000\255\255\
    ";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\068\000\068\000\068\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\086\000\086\000\086\000\
    \255\255\086\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \228\001\000\000\228\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\063\004\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\177\005\177\005\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\027\008\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\044\008\
    \255\255\255\255\255\255\044\008\044\008\000\000\255\255\255\255\
    \049\008\000\000\064\008\255\255\255\255\255\255\000\000\049\008\
    \255\255\255\255\255\255\000\000\255\255\000\000\255\255\064\008\
    \000\000\000\000\064\008\071\008\000\000\000\000\108\008\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\085\008\000\000\000\000\000\000\255\255\
    \255\255\000\000\255\255\000\000\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\255\255\000\000\110\008\000\000\000\000\
    \000\000\114\008\000\000\000\000\000\000\118\008\000\000\000\000\
    \255\255\255\255\000\000\000\000\125\008\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \000\000\138\008\000\000\255\255\000\000\142\008\000\000\000\000\
    \145\008\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\182\008\000\000\182\008\182\008\187\008\
    \182\008\182\008\000\000\187\008\000\000\190\008\000\000\000\000\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\000\000\213\008\000\000\255\255\000\000\
    \255\255\000\000\255\255\255\255\000\000\222\008\000\000\255\255\
    \225\008\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\000\000\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\045\009\000\000\000\000\000\000\
    \049\009\000\000\255\255\000\000\053\009\000\000\255\255\000\000\
    \057\009\000\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \073\009\000\000\000\000\255\255\255\255\255\255\079\009\000\000\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\092\009\000\000\092\009\092\009\092\009\092\009\
    \098\009\000\000\098\009\098\009\101\009\101\009\101\009\000\000\
    \101\009\101\009\101\009\101\009\000\000\101\009\101\009\101\009\
    \113\009\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \121\009\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\166\009\000\000\167\009\167\009\
    ";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\026\000\024\000\024\000\024\000\025\000\048\000\024\000\
    \056\000\055\000\067\000\104\000\055\000\069\000\070\000\069\000\
    \070\000\076\000\085\000\085\000\088\000\111\000\088\000\088\000\
    \026\000\096\000\103\000\104\000\115\000\122\000\105\000\056\000\
    \023\000\067\000\050\000\050\000\069\000\070\000\069\000\070\000\
    \076\000\085\000\085\000\088\000\091\000\088\000\088\000\103\000\
    \096\000\103\000\110\000\115\000\087\000\087\000\089\000\087\000\
    \090\000\006\000\005\000\022\000\014\000\017\000\010\000\020\000\
    \016\000\019\000\121\000\067\002\021\000\004\000\008\000\015\000\
    \009\000\007\000\018\000\012\000\013\000\011\000\003\000\078\002\
    \143\002\077\003\135\003\094\004\066\002\087\004\093\004\091\004\
    \095\004\055\004\035\004\111\002\036\004\032\004\098\002\013\004\
    \010\002\011\001\007\002\055\004\011\002\011\001\097\002\051\004\
    \048\004\012\004\093\002\174\001\112\002\031\004\006\002\234\001\
    \034\004\110\002\096\002\008\002\009\002\158\001\235\001\067\001\
    \173\001\092\001\068\001\216\000\051\004\048\004\093\002\066\001\
    \236\001\069\001\190\000\159\001\218\000\217\000\032\004\233\001\
    \202\000\091\001\113\000\201\000\192\000\126\000\049\000\024\000\
    \125\000\191\000\048\000\035\004\124\000\036\004\031\004\013\004\
    \127\000\111\002\098\002\128\000\202\000\092\000\174\001\201\000\
    \092\001\012\004\097\002\158\001\010\002\049\000\007\002\071\000\
    \011\002\034\004\112\002\173\001\234\001\047\000\096\002\110\002\
    \091\001\159\001\006\002\235\001\060\000\061\000\067\001\008\002\
    \009\002\068\001\062\000\063\000\064\000\236\001\066\001\065\000\
    \069\001\066\000\072\000\073\000\233\001\074\000\030\000\029\000\
    \046\000\038\000\041\000\034\000\044\000\040\000\043\000\075\000\
    \078\000\045\000\028\000\032\000\039\000\033\000\031\000\042\000\
    \036\000\037\000\035\000\027\000\216\000\190\000\049\000\024\000\
    \053\000\079\000\048\000\059\000\053\000\218\000\217\000\192\000\
    \126\000\059\000\080\000\125\000\191\000\081\000\082\000\124\000\
    \076\000\083\000\084\000\127\000\093\000\049\000\128\000\053\000\
    \002\000\094\000\059\000\053\000\095\000\047\000\088\000\106\000\
    \059\000\054\000\099\000\100\000\255\255\255\255\255\255\076\000\
    \101\000\096\000\255\255\255\255\107\000\255\255\108\000\114\000\
    \117\000\118\000\119\000\178\000\169\000\088\000\030\000\029\000\
    \046\000\038\000\041\000\034\000\044\000\040\000\043\000\087\000\
    \096\000\045\000\028\000\032\000\039\000\033\000\031\000\042\000\
    \036\000\037\000\035\000\027\000\101\000\108\000\110\000\111\000\
    \115\000\184\000\112\000\173\000\119\000\185\000\153\000\147\000\
    \141\000\052\000\121\000\122\000\051\000\052\000\123\000\132\000\
    \051\000\174\000\131\000\101\000\108\000\110\000\135\000\115\000\
    \136\000\161\000\058\000\119\000\057\000\102\000\109\000\137\000\
    \058\000\121\000\057\000\160\000\138\000\120\000\139\000\077\000\
    \140\000\140\000\142\000\143\000\140\000\144\000\129\000\145\000\
    \146\000\148\000\149\000\150\000\151\000\152\000\134\000\130\000\
    \098\000\125\000\154\000\133\000\155\000\156\000\157\000\140\000\
    \158\000\159\000\125\000\097\000\167\000\162\000\163\000\164\000\
    \165\000\166\000\125\000\168\000\125\000\170\000\171\000\172\000\
    \125\000\177\000\175\000\176\000\125\000\179\000\180\000\181\000\
    \182\000\183\000\125\000\187\000\186\000\125\000\188\000\189\000\
    \125\000\199\000\198\000\193\000\176\000\194\000\195\000\197\000\
    \125\000\125\000\116\000\176\000\200\000\146\000\215\000\196\000\
    \203\000\204\000\205\000\206\000\207\000\207\000\209\000\210\000\
    \211\000\212\000\213\000\214\000\125\000\146\000\006\001\237\000\
    \239\000\219\000\236\000\221\000\235\000\222\000\220\000\223\000\
    \224\000\225\000\226\000\207\000\207\000\228\000\226\000\238\000\
    \229\000\230\000\231\000\232\000\233\000\234\000\176\000\176\000\
    \004\001\252\000\245\000\240\000\241\000\242\000\246\000\243\000\
    \244\000\226\000\125\000\250\000\247\000\226\000\248\000\249\000\
    \125\000\251\000\125\000\253\000\254\000\255\000\002\001\003\001\
    \000\001\000\001\125\000\005\001\007\001\208\000\008\001\009\001\
    \010\001\125\000\033\001\050\001\016\001\020\001\017\001\013\001\
    \032\001\018\001\176\000\022\001\021\001\255\255\227\000\000\001\
    \000\001\014\001\051\001\019\001\146\000\015\001\023\001\012\001\
    \024\001\024\001\026\001\027\001\028\001\029\001\030\001\031\001\
    \125\000\035\001\034\001\125\000\036\001\037\001\038\001\039\001\
    \040\001\040\001\042\001\043\001\044\001\045\001\046\001\024\001\
    \024\001\047\001\048\001\049\001\125\000\054\001\052\001\125\000\
    \053\001\125\000\055\001\056\001\057\001\058\001\058\001\040\001\
    \040\001\060\001\061\001\062\001\001\001\063\001\064\001\065\001\
    \005\001\090\001\125\000\084\001\071\001\078\001\075\001\073\001\
    \074\001\176\000\076\001\077\001\058\001\058\001\125\000\079\001\
    \080\001\081\001\082\001\072\001\025\001\083\001\125\000\070\001\
    \085\001\086\001\087\001\088\001\089\001\041\001\125\000\125\000\
    \150\001\093\001\094\001\095\001\140\001\095\001\136\001\132\001\
    \123\001\116\001\107\001\103\001\104\001\105\001\106\001\176\000\
    \108\001\113\001\110\001\111\001\109\001\112\001\176\000\114\001\
    \115\001\176\000\095\001\059\001\095\001\117\001\118\001\119\001\
    \120\001\121\001\122\001\176\000\124\001\125\001\126\001\127\001\
    \128\001\129\001\130\001\131\001\176\000\133\001\134\001\135\001\
    \176\000\137\001\138\001\139\001\176\000\141\001\142\001\143\001\
    \144\001\145\001\146\001\147\001\148\001\149\001\176\000\096\001\
    \151\001\102\001\152\001\153\001\154\001\100\001\155\001\125\000\
    \156\001\125\000\164\001\098\001\157\001\162\001\101\001\160\001\
    \099\001\097\001\176\000\161\001\163\001\176\000\165\001\166\001\
    \167\001\168\001\169\001\170\001\171\001\172\001\232\001\176\001\
    \177\001\224\001\219\001\175\001\182\001\179\001\180\001\181\001\
    \176\000\178\001\183\001\184\001\185\001\186\001\186\001\213\001\
    \210\001\203\001\201\001\192\001\193\001\199\001\195\001\196\001\
    \197\001\198\001\176\000\200\001\125\000\202\001\176\000\204\001\
    \205\001\206\001\194\001\206\001\186\001\186\001\208\001\209\001\
    \176\000\211\001\212\001\205\001\214\001\215\001\216\001\217\001\
    \218\001\125\000\220\001\221\001\221\001\223\001\176\000\225\001\
    \206\001\229\001\206\001\227\001\228\001\176\000\230\001\226\001\
    \231\001\125\000\125\000\005\002\249\001\245\001\237\001\238\001\
    \239\001\240\001\221\001\221\001\241\001\242\001\187\001\243\001\
    \244\001\189\001\191\001\176\000\246\001\247\001\248\001\188\001\
    \190\001\146\000\250\001\251\001\252\001\253\001\253\001\255\001\
    \000\002\001\002\002\002\003\002\004\002\125\000\125\000\091\002\
    \088\002\082\002\077\002\076\002\012\002\207\001\013\002\063\002\
    \060\002\015\002\015\002\044\002\253\001\253\001\018\002\019\002\
    \020\002\021\002\222\001\022\002\023\002\024\002\025\002\014\002\
    \025\002\043\002\036\002\029\002\030\002\031\002\032\002\033\002\
    \015\002\015\002\034\002\035\002\176\000\037\002\038\002\039\002\
    \040\002\041\002\042\002\176\000\176\000\025\002\045\002\025\002\
    \046\002\047\002\048\002\049\002\050\002\051\002\051\002\053\002\
    \054\002\055\002\056\002\254\001\057\002\058\002\059\002\176\000\
    \061\002\062\002\125\000\064\002\065\002\071\002\068\002\069\002\
    \070\002\125\000\016\002\072\002\051\002\051\002\073\002\074\002\
    \026\002\017\002\028\002\075\002\125\000\228\001\079\002\080\002\
    \081\002\125\000\083\002\084\002\085\002\086\002\087\002\125\000\
    \089\002\090\002\125\000\092\002\027\002\172\001\094\002\095\002\
    \014\002\105\002\100\002\099\002\146\000\101\002\052\002\102\002\
    \103\002\104\002\125\000\106\002\107\002\108\002\109\002\125\000\
    \125\000\006\004\113\002\005\004\115\002\146\002\138\002\128\002\
    \114\002\119\002\126\002\121\002\117\002\118\002\122\002\123\002\
    \124\002\116\002\125\002\120\002\146\000\127\002\146\000\130\002\
    \130\002\135\002\133\002\134\002\129\002\136\002\137\002\129\002\
    \139\002\140\002\141\002\142\002\144\002\145\002\125\000\147\002\
    \148\002\149\002\149\002\003\004\252\003\248\003\130\002\130\002\
    \244\003\236\003\227\003\237\003\188\003\192\003\103\003\131\003\
    \087\003\130\003\057\003\088\003\129\002\129\002\191\003\193\003\
    \149\002\149\002\167\003\102\003\219\002\168\003\196\002\238\003\
    \132\003\133\003\166\003\058\003\169\003\184\002\003\003\176\002\
    \056\003\172\001\172\002\198\002\173\002\002\003\174\002\207\002\
    \175\002\004\003\172\001\177\002\005\003\197\002\178\002\206\002\
    \179\002\180\002\154\002\157\002\163\002\160\002\158\002\165\002\
    \161\002\155\002\156\002\205\002\181\002\166\002\153\002\169\002\
    \152\002\162\002\150\002\164\002\159\002\167\002\171\002\151\002\
    \182\002\183\002\172\001\185\002\186\002\187\002\188\002\188\002\
    \170\002\190\002\191\002\255\255\131\002\255\255\132\002\168\002\
    \192\002\193\002\194\002\195\002\172\001\172\001\202\002\200\002\
    \201\002\199\002\172\001\203\002\172\001\188\002\188\002\204\002\
    \172\001\172\001\214\002\209\002\208\002\172\001\210\002\211\002\
    \212\002\213\002\172\001\215\002\216\002\217\002\218\002\172\001\
    \221\002\237\002\254\002\226\002\224\002\225\002\172\001\236\002\
    \227\002\228\002\222\002\228\002\230\002\231\002\232\002\233\002\
    \220\002\255\002\223\002\234\002\235\002\172\001\239\002\238\002\
    \172\001\240\002\241\002\242\002\189\002\243\002\244\002\244\002\
    \228\002\246\002\228\002\247\002\248\002\249\002\250\002\251\002\
    \252\002\253\002\172\001\172\001\000\003\001\003\172\001\050\003\
    \044\003\042\003\038\003\051\003\009\003\244\002\244\002\008\003\
    \030\003\022\003\016\003\011\003\012\003\013\003\014\003\015\003\
    \172\001\017\003\029\003\018\003\019\003\020\003\021\003\172\001\
    \023\003\024\003\025\003\026\003\027\003\028\003\229\002\172\001\
    \036\003\031\003\032\003\006\003\033\003\034\003\035\003\172\001\
    \037\003\172\001\039\003\245\002\007\003\040\003\041\003\172\001\
    \010\003\043\003\172\001\045\003\046\003\047\003\255\255\255\255\
    \048\003\049\003\172\001\053\003\052\003\172\001\054\003\055\003\
    \172\001\172\001\081\003\059\003\080\003\072\003\062\003\063\003\
    \070\003\060\003\061\003\065\003\066\003\067\003\068\003\069\003\
    \172\001\064\003\071\003\172\001\073\003\074\003\075\003\076\003\
    \078\003\079\003\172\001\172\001\082\003\083\003\084\003\085\003\
    \086\003\172\001\090\003\089\003\172\001\091\003\092\003\093\003\
    \094\003\094\003\096\003\097\003\098\003\099\003\100\003\101\003\
    \172\001\129\003\104\003\112\003\106\003\105\003\107\003\108\003\
    \109\003\110\003\111\003\172\001\113\003\114\003\115\003\094\003\
    \094\003\116\003\116\003\125\003\119\003\120\003\121\003\122\003\
    \123\003\124\003\172\001\126\003\127\003\128\003\172\001\172\001\
    \148\003\145\003\149\003\139\003\134\003\136\003\137\003\138\003\
    \116\003\116\003\172\001\140\003\141\003\142\003\143\003\144\003\
    \172\001\146\003\147\003\172\001\153\003\150\003\151\003\152\003\
    \172\001\095\003\154\003\155\003\157\003\161\003\158\003\159\003\
    \160\003\172\001\162\003\163\003\164\003\165\003\172\001\187\003\
    \172\001\181\003\171\003\175\003\172\003\156\003\173\003\174\003\
    \172\001\176\003\118\003\177\003\178\003\179\003\117\003\180\003\
    \172\001\182\003\183\003\184\003\185\003\170\003\186\003\172\001\
    \172\001\189\003\190\003\172\001\222\003\217\003\194\003\196\003\
    \209\003\202\003\197\003\198\003\199\003\203\003\200\003\201\003\
    \172\001\207\003\204\003\205\003\206\003\172\001\195\003\208\003\
    \172\001\210\003\211\003\212\003\213\003\213\003\215\003\216\003\
    \172\001\218\003\219\003\220\003\221\003\172\001\223\003\224\003\
    \225\003\226\003\172\001\228\003\229\003\230\003\231\003\232\003\
    \233\003\234\003\172\001\213\003\213\003\235\003\172\001\242\003\
    \172\001\239\003\240\003\241\003\172\001\243\003\172\001\245\003\
    \246\003\247\003\172\001\249\003\250\003\251\003\172\001\253\003\
    \254\003\255\003\000\004\001\004\002\004\172\001\004\004\172\001\
    \125\000\007\004\008\004\009\004\010\004\011\004\125\000\026\004\
    \014\004\015\004\016\004\016\004\018\004\019\004\020\004\021\004\
    \022\004\023\004\024\004\025\004\129\002\027\004\028\004\029\004\
    \214\003\030\004\005\001\033\004\228\001\125\000\045\004\042\004\
    \125\000\016\004\016\004\038\004\039\004\040\004\041\004\228\001\
    \044\004\037\004\125\000\043\004\228\001\046\004\047\004\125\000\
    \049\004\050\004\125\000\052\004\053\004\054\004\125\000\056\004\
    \057\004\058\004\059\004\060\004\061\004\125\000\064\004\087\004\
    \064\004\019\008\088\004\015\008\012\008\253\007\241\007\042\006\
    \254\007\226\007\066\006\227\007\223\007\160\004\240\007\046\006\
    \215\005\017\004\212\005\051\005\216\005\064\004\145\004\064\004\
    \123\005\045\006\239\007\067\006\222\007\086\004\211\005\225\007\
    \065\006\052\005\241\004\213\005\214\005\122\005\067\005\100\004\
    \090\004\127\004\117\004\216\004\102\004\066\005\217\004\101\004\
    \103\004\069\005\240\004\215\004\068\005\218\004\070\004\067\004\
    \078\004\077\004\081\004\069\004\068\004\080\004\084\004\090\004\
    \104\004\083\004\066\004\073\004\079\004\072\004\071\004\082\004\
    \075\004\076\004\074\004\065\004\183\005\150\004\105\004\097\004\
    \090\004\091\004\106\004\184\005\092\004\107\004\107\004\151\004\
    \099\004\098\004\090\004\091\004\152\004\185\005\092\004\096\004\
    \096\004\109\004\110\004\096\004\182\005\119\004\121\004\090\004\
    \118\004\111\004\112\004\113\004\107\004\107\004\114\004\115\004\
    \116\004\090\004\116\004\143\004\135\004\120\004\096\004\128\004\
    \122\004\123\004\124\004\129\004\089\004\125\004\126\004\127\004\
    \133\004\130\004\131\004\132\004\127\004\134\004\127\004\136\004\
    \137\004\138\004\139\004\139\004\141\004\142\004\108\004\127\004\
    \144\004\146\004\147\004\148\004\149\004\127\004\158\004\155\004\
    \153\004\154\004\116\004\156\004\116\004\127\004\157\004\127\004\
    \159\004\139\004\139\004\127\004\089\004\162\004\182\004\199\004\
    \165\004\169\004\166\004\167\004\181\004\116\004\089\004\163\004\
    \171\004\170\004\127\004\164\004\172\004\161\004\200\004\168\004\
    \173\004\173\004\175\004\176\004\177\004\178\004\179\004\180\004\
    \127\004\184\004\183\004\127\004\185\004\186\004\187\004\188\004\
    \189\004\189\004\191\004\192\004\193\004\194\004\195\004\173\004\
    \173\004\196\004\197\004\198\004\127\004\203\004\140\004\127\004\
    \201\004\202\004\127\004\204\004\205\004\206\004\209\004\189\004\
    \189\004\207\004\207\004\210\004\211\004\212\004\213\004\214\004\
    \144\004\239\004\127\004\233\004\220\004\227\004\224\004\222\004\
    \223\004\116\004\225\004\226\004\127\004\085\004\228\004\229\004\
    \207\004\207\004\230\004\221\004\174\004\231\004\232\004\219\004\
    \127\004\234\004\235\004\236\004\237\004\190\004\238\004\127\004\
    \127\004\043\005\242\004\243\004\244\004\244\004\033\005\029\005\
    \025\005\016\005\009\005\000\005\252\004\253\004\254\004\255\004\
    \116\004\001\005\006\005\003\005\004\005\002\005\005\005\116\004\
    \007\005\008\005\116\004\244\004\244\004\010\005\011\005\208\004\
    \012\005\013\005\014\005\015\005\116\004\017\005\018\005\019\005\
    \020\005\021\005\022\005\023\005\024\005\116\004\026\005\027\005\
    \028\005\116\004\030\005\031\005\032\005\116\004\034\005\035\005\
    \036\005\037\005\038\005\039\005\040\005\041\005\042\005\245\004\
    \116\004\251\004\044\005\045\005\046\005\249\004\047\005\048\005\
    \049\005\127\004\127\004\247\004\050\005\057\005\250\004\053\005\
    \248\004\246\004\055\005\054\005\116\004\056\005\116\004\058\005\
    \059\005\060\005\061\005\062\005\063\005\064\005\065\005\116\005\
    \110\005\077\005\070\005\117\005\076\005\073\005\072\005\116\004\
    \127\004\106\005\098\005\090\005\084\005\079\005\080\005\081\005\
    \071\005\082\005\083\005\127\004\097\005\085\005\086\005\087\005\
    \088\005\089\005\127\004\091\005\092\005\093\005\094\005\095\005\
    \074\005\096\005\127\004\104\005\099\005\100\005\101\005\102\005\
    \103\005\075\005\127\004\105\005\127\004\078\005\107\005\108\005\
    \109\005\127\004\111\005\112\005\113\005\114\005\115\005\127\004\
    \119\005\118\005\127\004\120\005\121\005\127\004\181\005\125\005\
    \126\005\173\005\168\005\124\005\131\005\128\005\129\005\130\005\
    \116\004\127\005\132\005\133\005\134\005\135\005\135\005\162\005\
    \159\005\152\005\150\005\141\005\142\005\148\005\144\005\145\005\
    \146\005\147\005\116\004\149\005\127\004\151\005\116\004\153\005\
    \154\005\155\005\143\005\155\005\135\005\135\005\157\005\158\005\
    \116\004\160\005\161\005\154\005\163\005\164\005\165\005\166\005\
    \167\005\127\004\169\005\170\005\170\005\172\005\116\004\174\005\
    \155\005\178\005\155\005\176\005\177\005\116\004\179\005\175\005\
    \180\005\127\004\127\004\210\005\198\005\194\005\186\005\187\005\
    \188\005\189\005\170\005\170\005\190\005\191\005\136\005\192\005\
    \193\005\138\005\140\005\116\004\195\005\196\005\197\005\137\005\
    \139\005\127\004\199\005\200\005\201\005\202\005\202\005\204\005\
    \205\005\206\005\207\005\208\005\209\005\127\004\127\004\040\006\
    \037\006\031\006\026\006\025\006\217\005\156\005\218\005\012\006\
    \009\006\220\005\220\005\249\005\202\005\202\005\223\005\224\005\
    \225\005\226\005\171\005\227\005\228\005\229\005\230\005\219\005\
    \230\005\248\005\241\005\234\005\235\005\236\005\237\005\238\005\
    \220\005\220\005\239\005\240\005\116\004\242\005\243\005\244\005\
    \245\005\246\005\247\005\116\004\116\004\230\005\250\005\230\005\
    \251\005\252\005\253\005\254\005\255\005\000\006\000\006\002\006\
    \003\006\004\006\005\006\203\005\006\006\007\006\008\006\116\004\
    \010\006\011\006\127\004\013\006\014\006\016\006\020\006\017\006\
    \018\006\019\006\221\005\127\004\000\006\000\006\021\006\022\006\
    \231\005\222\005\233\005\023\006\024\006\127\004\015\006\177\005\
    \027\006\028\006\029\006\030\006\127\004\032\006\033\006\034\006\
    \035\006\036\006\127\004\038\006\232\005\039\006\127\004\041\006\
    \065\005\043\006\044\006\219\005\060\006\047\006\001\006\048\006\
    \049\006\049\006\051\006\052\006\053\006\054\006\055\006\056\006\
    \057\006\058\006\059\006\061\006\062\006\063\006\064\006\144\004\
    \127\004\216\007\215\007\068\006\070\006\100\006\092\006\049\006\
    \049\006\069\006\083\006\074\006\073\006\072\006\075\006\075\006\
    \080\006\071\006\078\006\079\006\084\006\059\006\081\006\082\006\
    \059\006\090\006\085\006\086\006\087\006\088\006\089\006\127\004\
    \091\006\127\004\093\006\094\006\095\006\075\006\075\006\096\006\
    \097\006\098\006\099\006\127\004\101\006\102\006\103\006\213\007\
    \206\007\202\007\198\007\059\006\059\006\181\007\190\007\050\006\
    \191\007\146\007\142\007\057\007\085\007\041\007\084\007\011\007\
    \042\007\121\007\145\007\147\007\122\007\103\006\103\006\173\006\
    \056\007\120\007\161\006\123\007\192\007\086\007\087\007\152\006\
    \012\007\213\006\160\006\123\006\065\005\010\007\150\006\138\006\
    \212\006\151\006\125\006\130\006\214\006\103\006\159\006\215\006\
    \126\006\127\006\128\006\129\006\124\006\065\005\131\006\132\006\
    \133\006\134\006\135\006\122\006\136\006\137\006\065\005\139\006\
    \140\006\141\006\144\006\145\006\142\006\146\006\147\006\148\006\
    \149\006\065\005\255\255\076\006\255\255\077\006\108\006\111\006\
    \117\006\114\006\112\006\119\006\115\006\109\006\110\006\142\006\
    \065\005\120\006\107\006\142\006\106\006\116\006\104\006\118\006\
    \113\006\121\006\156\006\105\006\155\006\153\006\154\006\065\005\
    \157\006\065\005\158\006\065\005\065\005\168\006\142\006\163\006\
    \162\006\065\005\164\006\165\006\166\006\167\006\065\005\169\006\
    \170\006\171\006\172\006\065\005\175\006\191\006\208\006\180\006\
    \178\006\179\006\065\005\190\006\181\006\184\006\176\006\182\006\
    \182\006\185\006\186\006\187\006\174\006\209\006\177\006\188\006\
    \189\006\065\005\193\006\192\006\065\005\194\006\195\006\196\006\
    \197\006\198\006\200\006\198\006\143\006\201\006\182\006\182\006\
    \202\006\203\006\204\006\205\006\206\006\207\006\065\005\065\005\
    \210\006\211\006\065\005\004\007\254\006\252\006\248\006\005\007\
    \198\006\219\006\198\006\240\006\218\006\232\006\226\006\221\006\
    \222\006\223\006\224\006\225\006\065\005\239\006\227\006\228\006\
    \229\006\230\006\231\006\065\005\233\006\234\006\235\006\236\006\
    \237\006\238\006\065\005\183\006\246\006\241\006\242\006\243\006\
    \216\006\244\006\245\006\065\005\247\006\255\255\255\255\199\006\
    \065\005\217\006\249\006\250\006\251\006\220\006\065\005\253\006\
    \065\005\255\006\000\007\001\007\002\007\003\007\065\005\007\007\
    \006\007\065\005\008\007\009\007\065\005\065\005\035\007\013\007\
    \034\007\026\007\016\007\017\007\024\007\014\007\015\007\019\007\
    \020\007\021\007\022\007\023\007\065\005\018\007\025\007\065\005\
    \027\007\028\007\029\007\030\007\031\007\032\007\033\007\065\005\
    \065\005\036\007\037\007\038\007\039\007\040\007\065\005\044\007\
    \043\007\065\005\045\007\046\007\047\007\048\007\048\007\050\007\
    \051\007\052\007\053\007\054\007\055\007\065\005\083\007\058\007\
    \066\007\060\007\059\007\061\007\062\007\063\007\064\007\065\007\
    \065\005\067\007\068\007\069\007\048\007\048\007\070\007\070\007\
    \079\007\073\007\074\007\075\007\076\007\077\007\078\007\065\005\
    \080\007\081\007\082\007\065\005\065\005\102\007\099\007\103\007\
    \093\007\088\007\089\007\090\007\091\007\070\007\070\007\092\007\
    \065\005\094\007\095\007\096\007\097\007\098\007\065\005\100\007\
    \101\007\065\005\107\007\104\007\105\007\106\007\049\007\065\005\
    \108\007\109\007\111\007\115\007\112\007\113\007\114\007\065\005\
    \116\007\117\007\118\007\119\007\065\005\141\007\065\005\135\007\
    \125\007\129\007\126\007\110\007\127\007\128\007\065\005\072\007\
    \130\007\131\007\132\007\071\007\133\007\134\007\065\005\136\007\
    \137\007\138\007\139\007\124\007\140\007\065\005\065\005\143\007\
    \144\007\065\005\176\007\171\007\148\007\150\007\163\007\156\007\
    \151\007\152\007\153\007\157\007\154\007\155\007\065\005\161\007\
    \158\007\159\007\160\007\065\005\149\007\162\007\065\005\164\007\
    \165\007\166\007\167\007\167\007\169\007\170\007\065\005\172\007\
    \173\007\174\007\175\007\065\005\177\007\178\007\179\007\180\007\
    \065\005\182\007\183\007\184\007\185\007\186\007\187\007\188\007\
    \065\005\167\007\167\007\189\007\065\005\196\007\065\005\193\007\
    \194\007\195\007\065\005\197\007\065\005\199\007\200\007\201\007\
    \065\005\203\007\204\007\205\007\065\005\207\007\208\007\209\007\
    \210\007\211\007\212\007\065\005\214\007\065\005\127\004\217\007\
    \218\007\219\007\220\007\221\007\127\004\224\007\177\005\127\004\
    \236\007\233\007\127\004\229\007\230\007\231\007\232\007\177\005\
    \127\004\235\007\177\005\228\007\234\007\237\007\168\007\238\007\
    \127\004\248\007\243\007\242\007\127\004\244\007\245\007\246\007\
    \247\007\127\004\249\007\250\007\251\007\252\007\127\004\000\008\
    \255\007\127\004\001\008\002\008\003\008\004\008\004\008\006\008\
    \007\008\008\008\009\008\010\008\011\008\127\004\013\008\014\008\
    \127\004\016\008\017\008\018\008\127\004\020\008\021\008\022\008\
    \023\008\024\008\025\008\127\004\004\008\004\008\035\008\032\008\
    \047\008\031\008\033\008\046\008\032\008\046\008\032\008\036\008\
    \063\008\046\008\039\008\057\008\038\008\032\008\046\008\032\008\
    \036\008\056\008\046\008\039\008\059\008\035\008\061\008\038\008\
    \032\008\255\255\046\008\036\008\046\008\045\008\083\008\034\008\
    \039\008\039\008\031\008\038\008\083\008\046\008\005\008\031\008\
    \039\008\031\008\039\008\040\008\039\008\037\008\038\008\047\008\
    \031\008\030\008\031\008\106\008\040\008\039\008\255\255\039\008\
    \037\008\043\008\066\008\031\008\255\255\043\008\056\008\039\008\
    \107\008\041\008\105\008\039\008\046\008\046\008\047\008\104\008\
    \046\008\042\008\055\008\050\008\039\008\066\008\051\008\055\008\
    \050\008\066\008\066\008\051\008\028\008\056\008\043\008\059\008\
    \094\008\061\008\060\008\046\008\062\008\083\008\065\008\083\008\
    \087\008\055\008\086\008\083\008\066\008\089\008\055\008\091\008\
    \031\008\066\008\092\008\091\008\100\008\098\008\054\008\083\008\
    \099\008\065\008\102\008\054\008\095\008\096\008\065\008\081\008\
    \084\008\093\008\082\008\103\008\083\008\053\008\095\008\096\008\
    \112\008\123\008\083\008\100\008\122\008\136\008\131\008\130\008\
    \131\008\132\008\133\008\097\008\140\008\083\008\116\008\072\008\
    \074\008\076\008\075\008\152\008\068\008\078\008\139\008\077\008\
    \089\008\143\008\159\008\090\008\052\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\158\008\
    \058\008\073\008\083\008\152\008\152\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \080\008\191\008\068\008\211\008\070\008\083\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \068\008\068\008\068\008\068\008\068\008\068\008\068\008\068\008\
    \079\008\088\008\101\008\101\008\101\008\098\008\101\008\107\008\
    \101\008\210\008\209\008\100\008\098\008\029\008\193\008\099\008\
    \206\008\101\008\101\008\202\008\201\008\101\008\135\008\133\008\
    \255\255\101\008\134\008\101\008\203\008\121\008\107\008\207\008\
    \120\008\097\008\100\008\097\008\185\008\193\008\128\008\255\255\
    \101\008\129\008\097\008\255\255\220\008\135\008\135\008\133\008\
    \097\008\255\255\134\008\217\008\255\255\146\008\146\008\177\008\
    \178\008\146\008\176\008\175\008\168\008\169\008\207\008\167\008\
    \166\008\208\008\233\008\004\009\255\255\135\008\223\008\177\008\
    \178\008\255\255\176\008\175\008\146\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\255\255\127\008\248\008\255\255\223\008\255\255\239\008\
    \245\008\255\255\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\186\008\255\255\186\008\214\008\
    \243\008\246\008\249\008\000\009\111\008\216\008\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \253\008\115\008\001\009\002\009\005\009\069\008\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\233\008\255\255\026\009\255\255\026\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\205\008\203\008\042\009\043\009\204\008\041\009\
    \040\009\051\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\019\009\020\009\055\009\018\009\
    \017\009\205\008\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \060\009\255\255\119\008\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\146\008\146\008\062\009\
    \242\008\146\008\063\009\066\009\126\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\067\009\064\009\068\009\146\008\188\008\188\008\242\008\
    \070\009\205\008\203\008\059\009\155\008\204\008\154\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\064\009\076\009\074\009\188\008\086\009\255\255\215\008\
    \205\008\081\009\255\255\081\009\255\255\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\156\008\
    \087\009\157\008\088\009\152\008\241\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\115\009\
    \116\009\131\009\255\255\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\074\009\131\009\146\009\077\009\151\008\150\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\149\008\148\008\146\009\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\147\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\153\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\188\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\000\000\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\219\008\217\008\
    \058\009\000\000\218\008\075\009\160\008\114\009\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \038\009\039\009\050\009\037\009\036\009\219\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\088\009\000\000\000\000\089\009\152\008\000\000\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\000\000\000\000\000\000\000\000\152\008\
    \000\000\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\219\008\217\008\034\009\035\009\218\008\
    \033\009\032\009\091\009\162\009\163\009\093\009\161\009\160\009\
    \165\008\164\008\139\009\140\009\000\000\138\009\137\009\245\008\
    \246\008\000\000\219\008\247\008\163\008\162\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\245\008\000\000\
    \161\008\248\008\249\008\000\000\000\000\250\008\001\009\002\009\
    \006\009\006\009\003\009\119\009\006\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\000\000\
    \248\008\000\000\174\008\173\008\000\000\001\009\000\000\006\009\
    \059\009\074\009\119\009\000\000\000\000\255\255\172\008\171\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \255\255\000\000\170\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\054\009\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \000\000\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \255\255\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\255\255\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\000\000\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\000\000\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\184\008\
    \184\008\000\000\223\008\184\008\184\008\184\008\165\009\060\009\
    \184\008\255\255\255\255\193\008\191\008\091\009\165\009\192\008\
    \093\009\255\008\253\008\165\009\068\009\254\008\184\008\097\009\
    \183\008\223\008\099\009\184\008\097\009\255\255\060\009\099\009\
    \255\255\000\000\193\008\000\000\181\008\195\008\197\008\196\008\
    \255\008\185\008\200\008\068\009\199\008\226\008\239\008\158\009\
    \159\009\240\008\157\009\156\009\000\000\069\009\251\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\000\000\226\008\000\000\227\008\070\009\
    \059\009\085\009\081\009\071\009\235\008\084\009\000\000\000\000\
    \000\000\000\000\236\008\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\198\008\070\009\000\000\
    \085\009\000\000\000\000\061\009\000\000\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\000\000\
    \000\000\238\008\000\000\233\008\000\000\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\234\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\117\009\000\000\
    \108\009\255\008\253\008\107\009\000\000\254\008\232\008\231\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\230\008\229\008\119\009\117\009\180\008\128\009\
    \255\008\255\255\064\009\255\255\255\255\000\000\000\000\116\009\
    \255\255\000\000\194\008\255\255\255\255\000\000\228\008\000\000\
    \252\008\255\255\000\000\119\009\000\000\255\255\128\009\000\000\
    \000\000\064\009\255\255\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\237\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\065\009\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\000\000\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\242\008\243\008\000\000\000\000\244\008\
    \000\000\007\009\000\000\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\154\009\155\009\000\000\
    \153\009\152\009\242\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\000\000\000\000\
    \000\000\000\000\233\008\000\000\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\255\255\241\008\
    \252\008\242\008\243\008\000\000\000\000\244\008\000\000\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \242\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\000\000\000\000\000\000\012\009\011\009\000\000\
    \000\000\047\009\085\009\081\009\000\000\233\008\084\009\000\000\
    \000\000\010\009\009\009\000\000\007\009\241\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \047\009\085\009\046\009\000\000\000\000\008\009\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\000\000\000\000\000\000\000\000\233\008\000\000\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\013\009\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\000\000\000\000\103\009\000\000\026\009\
    \107\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\000\000\
    \012\009\011\009\000\000\000\000\000\000\000\000\000\000\000\000\
    \103\009\000\000\000\000\107\009\010\009\009\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\117\009\000\000\
    \008\009\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\000\000\000\000\000\000\000\000\117\009\000\000\000\000\
    \000\000\000\000\025\009\024\009\000\000\000\000\000\000\118\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\023\009\022\009\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\000\000\000\000\000\000\000\000\
    \255\255\000\000\021\009\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\000\000\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\000\000\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\255\255\000\000\000\000\000\000\000\000\
    \000\000\007\009\000\000\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\000\000\000\000\000\000\
    \000\000\000\000\000\000\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\255\255\000\000\
    \000\000\000\000\233\008\000\000\233\008\233\008\233\008\233\008\
    \014\009\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\012\009\011\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\233\008\000\000\000\000\
    \000\000\010\009\009\009\000\000\007\009\000\000\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \000\000\000\000\000\000\000\000\000\000\008\009\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\108\009\000\000\000\000\107\009\233\008\000\000\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \015\009\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \083\009\081\009\108\009\091\009\082\009\107\009\093\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\083\009\
    \012\009\011\009\000\000\000\000\000\000\000\000\000\000\000\000\
    \233\008\000\000\000\000\000\000\010\009\009\009\095\009\007\009\
    \000\000\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\000\000\000\000\000\000\000\000\000\000\
    \008\009\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\080\009\109\009\000\000\000\000\
    \233\008\000\000\233\008\233\008\233\008\233\008\016\009\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\000\000\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\000\000\000\000\103\009\091\009\255\255\
    \111\009\093\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\000\000\000\000\
    \000\000\000\000\000\000\012\009\011\009\000\000\000\000\000\000\
    \000\000\000\000\000\000\233\008\000\000\000\000\000\000\010\009\
    \009\009\096\009\007\009\000\000\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\000\000\000\000\
    \000\000\000\000\000\000\008\009\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\255\255\
    \255\255\255\255\000\000\233\008\000\000\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\000\000\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\000\000\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\012\009\011\009\
    \103\009\000\000\000\000\107\009\000\000\000\000\026\009\000\000\
    \000\000\000\000\010\009\009\009\000\000\007\009\000\000\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\000\000\000\000\000\000\000\000\000\000\008\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\000\000\255\255\255\255\000\000\026\009\000\000\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\000\000\
    \000\000\031\009\030\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\029\009\028\009\000\000\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\027\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\255\255\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\000\000\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\090\009\
    \091\009\097\009\129\009\093\009\099\009\105\009\103\009\105\009\
    \103\009\104\009\108\009\106\009\000\000\107\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\090\009\000\000\
    \000\000\129\009\000\000\000\000\105\009\000\000\105\009\000\000\
    \000\000\000\000\000\000\094\009\096\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\000\000\000\000\128\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \000\000\102\009\000\000\102\009\127\009\110\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\126\009\
    \125\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\124\009\123\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\000\000\000\000\122\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\000\000\
    \255\255\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\000\000\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\000\000\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\129\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\000\000\000\000\129\009\000\000\
    \000\000\000\000\000\000\000\000\000\000\131\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\130\009\000\000\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \128\009\000\000\000\000\000\000\000\000\000\000\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\000\000\000\000\000\000\000\000\131\009\000\000\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\000\000\000\000\000\000\000\000\146\009\
    \000\000\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \136\009\135\009\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\134\009\133\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \132\009\000\000\000\000\000\000\000\000\129\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\145\009\144\009\129\009\000\000\000\000\000\000\
    \000\000\000\000\000\000\131\009\000\000\000\000\143\009\142\009\
    \000\000\000\000\130\009\000\000\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\128\009\000\000\
    \000\000\000\000\141\009\000\000\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\000\000\
    \000\000\000\000\000\000\131\009\000\000\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\136\009\135\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\134\009\133\009\000\000\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\132\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\000\000\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\000\000\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\000\000\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\000\000\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\129\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \129\009\000\000\000\000\000\000\000\000\000\000\000\000\146\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\130\009\000\000\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\128\009\000\000\000\000\000\000\000\000\000\000\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\000\000\000\000\000\000\000\000\146\009\
    \000\000\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\151\009\150\009\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\149\009\148\009\
    \000\000\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\147\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\000\000\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\000\000\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\024\000\025\000\000\000\024\000\048\000\
    \054\000\054\000\066\000\105\000\054\000\067\000\068\000\069\000\
    \070\000\075\000\084\000\085\000\086\000\112\000\087\000\088\000\
    \000\000\095\000\103\000\103\000\114\000\123\000\103\000\054\000\
    \000\000\066\000\023\000\047\000\067\000\068\000\069\000\070\000\
    \075\000\084\000\085\000\086\000\090\000\087\000\088\000\102\000\
    \095\000\103\000\109\000\114\000\085\000\086\000\087\000\087\000\
    \088\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\120\000\065\002\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\077\002\
    \142\002\076\003\134\003\086\004\065\002\088\004\089\004\092\004\
    \094\004\003\000\006\000\009\000\006\000\007\000\010\000\008\000\
    \012\000\018\000\012\000\027\000\012\000\042\000\010\000\004\000\
    \005\000\008\000\011\000\014\000\009\000\007\000\012\000\013\000\
    \006\000\009\000\010\000\012\000\012\000\015\000\013\000\017\000\
    \014\000\016\000\017\000\019\000\028\000\029\000\035\000\017\000\
    \013\000\017\000\021\000\015\000\019\000\019\000\031\000\013\000\
    \020\000\016\000\051\000\020\000\021\000\022\000\026\000\026\000\
    \022\000\021\000\026\000\030\000\022\000\030\000\031\000\032\000\
    \022\000\033\000\034\000\022\000\044\000\052\000\038\000\044\000\
    \040\000\032\000\034\000\039\000\036\000\026\000\036\000\057\000\
    \036\000\030\000\033\000\038\000\037\000\026\000\034\000\033\000\
    \040\000\039\000\036\000\037\000\058\000\060\000\041\000\036\000\
    \036\000\041\000\061\000\062\000\063\000\037\000\041\000\064\000\
    \041\000\065\000\071\000\072\000\037\000\073\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\074\000\
    \077\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\043\000\045\000\049\000\049\000\
    \050\000\078\000\049\000\056\000\053\000\043\000\043\000\045\000\
    \046\000\059\000\079\000\046\000\045\000\080\000\081\000\046\000\
    \076\000\082\000\083\000\046\000\092\000\049\000\046\000\050\000\
    \000\000\093\000\056\000\053\000\094\000\049\000\089\000\097\000\
    \059\000\050\000\098\000\099\000\067\000\068\000\069\000\076\000\
    \100\000\096\000\085\000\086\000\106\000\087\000\107\000\113\000\
    \116\000\117\000\118\000\126\000\129\000\089\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\089\000\
    \096\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\101\000\108\000\110\000\110\000\
    \115\000\124\000\110\000\127\000\119\000\124\000\131\000\132\000\
    \133\000\050\000\121\000\121\000\050\000\053\000\121\000\128\000\
    \053\000\127\000\128\000\101\000\108\000\110\000\134\000\115\000\
    \135\000\130\000\056\000\119\000\056\000\101\000\108\000\136\000\
    \059\000\121\000\059\000\130\000\137\000\119\000\138\000\076\000\
    \139\000\139\000\141\000\142\000\139\000\143\000\128\000\144\000\
    \145\000\147\000\148\000\149\000\150\000\151\000\128\000\128\000\
    \096\000\152\000\153\000\128\000\154\000\155\000\156\000\139\000\
    \157\000\158\000\159\000\096\000\160\000\161\000\162\000\163\000\
    \164\000\165\000\166\000\167\000\168\000\169\000\170\000\171\000\
    \172\000\173\000\174\000\175\000\177\000\178\000\179\000\180\000\
    \181\000\182\000\183\000\184\000\185\000\186\000\187\000\188\000\
    \189\000\190\000\191\000\192\000\193\000\192\000\194\000\195\000\
    \190\000\196\000\115\000\198\000\199\000\200\000\201\000\195\000\
    \202\000\203\000\204\000\205\000\206\000\207\000\208\000\209\000\
    \210\000\211\000\212\000\213\000\214\000\215\000\216\000\217\000\
    \217\000\218\000\217\000\219\000\220\000\221\000\219\000\222\000\
    \223\000\224\000\225\000\206\000\207\000\227\000\226\000\217\000\
    \228\000\229\000\230\000\231\000\232\000\233\000\234\000\235\000\
    \236\000\237\000\238\000\239\000\240\000\241\000\238\000\242\000\
    \243\000\225\000\244\000\245\000\246\000\226\000\247\000\248\000\
    \249\000\250\000\251\000\252\000\253\000\254\000\001\001\002\001\
    \255\000\000\001\003\001\004\001\006\001\207\000\007\001\008\001\
    \009\001\010\001\013\001\012\001\015\001\014\001\016\001\011\001\
    \013\001\017\001\018\001\019\001\020\001\089\000\226\000\255\000\
    \000\001\011\001\012\001\014\001\021\001\011\001\022\001\011\001\
    \023\001\024\001\025\001\026\001\027\001\028\001\029\001\030\001\
    \031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
    \039\001\040\001\041\001\042\001\043\001\044\001\045\001\023\001\
    \024\001\046\001\047\001\048\001\049\001\050\001\051\001\050\001\
    \052\001\053\001\054\001\055\001\056\001\057\001\058\001\039\001\
    \040\001\059\001\060\001\061\001\000\001\062\001\063\001\064\001\
    \065\001\066\001\067\001\068\001\069\001\070\001\071\001\072\001\
    \073\001\074\001\075\001\076\001\057\001\058\001\077\001\078\001\
    \079\001\080\001\081\001\069\001\024\001\082\001\083\001\069\001\
    \084\001\085\001\086\001\087\001\088\001\040\001\089\001\090\001\
    \091\001\092\001\093\001\094\001\096\001\095\001\097\001\098\001\
    \099\001\100\001\101\001\102\001\103\001\104\001\105\001\106\001\
    \107\001\108\001\109\001\110\001\107\001\111\001\112\001\113\001\
    \114\001\115\001\094\001\058\001\095\001\116\001\117\001\118\001\
    \119\001\120\001\121\001\122\001\123\001\124\001\125\001\126\001\
    \127\001\128\001\129\001\130\001\131\001\132\001\133\001\134\001\
    \135\001\136\001\137\001\138\001\139\001\140\001\141\001\142\001\
    \143\001\144\001\145\001\146\001\147\001\148\001\149\001\095\001\
    \150\001\095\001\151\001\152\001\153\001\095\001\154\001\156\001\
    \155\001\157\001\158\001\095\001\155\001\160\001\095\001\159\001\
    \095\001\095\001\161\001\159\001\162\001\163\001\164\001\165\001\
    \166\001\167\001\168\001\169\001\170\001\171\001\173\001\174\001\
    \174\001\175\001\176\001\174\001\177\001\178\001\179\001\180\001\
    \181\001\174\001\182\001\183\001\184\001\185\001\186\001\187\001\
    \188\001\189\001\190\001\191\001\192\001\193\001\194\001\195\001\
    \196\001\197\001\198\001\199\001\200\001\201\001\202\001\203\001\
    \204\001\205\001\192\001\206\001\185\001\186\001\207\001\208\001\
    \209\001\210\001\211\001\212\001\213\001\214\001\215\001\216\001\
    \217\001\218\001\219\001\220\001\221\001\222\001\223\001\224\001\
    \205\001\226\001\206\001\225\001\227\001\222\001\229\001\225\001\
    \230\001\231\001\232\001\233\001\234\001\235\001\236\001\237\001\
    \238\001\239\001\220\001\221\001\240\001\241\001\186\001\242\001\
    \243\001\186\001\186\001\244\001\245\001\246\001\247\001\186\001\
    \186\001\248\001\249\001\250\001\251\001\252\001\253\001\254\001\
    \255\001\000\002\001\002\002\002\003\002\004\002\005\002\006\002\
    \007\002\008\002\009\002\010\002\011\002\206\001\011\002\012\002\
    \013\002\014\002\015\002\016\002\252\001\253\001\017\002\018\002\
    \019\002\020\002\221\001\021\002\022\002\023\002\024\002\011\002\
    \025\002\026\002\027\002\028\002\029\002\030\002\031\002\032\002\
    \014\002\015\002\033\002\034\002\035\002\036\002\037\002\038\002\
    \039\002\040\002\041\002\042\002\043\002\024\002\044\002\025\002\
    \045\002\046\002\047\002\048\002\049\002\050\002\051\002\052\002\
    \053\002\054\002\055\002\253\001\056\002\057\002\058\002\059\002\
    \060\002\061\002\062\002\063\002\064\002\066\002\067\002\068\002\
    \069\002\070\002\015\002\071\002\050\002\051\002\072\002\073\002\
    \025\002\015\002\025\002\074\002\075\002\076\002\078\002\079\002\
    \080\002\081\002\082\002\083\002\084\002\085\002\086\002\087\002\
    \088\002\089\002\090\002\091\002\025\002\092\002\093\002\094\002\
    \095\002\096\002\097\002\098\002\099\002\100\002\051\002\101\002\
    \102\002\103\002\104\002\105\002\106\002\107\002\108\002\109\002\
    \110\002\111\002\112\002\113\002\114\002\115\002\116\002\117\002\
    \112\002\118\002\119\002\120\002\114\002\114\002\121\002\122\002\
    \123\002\114\002\124\002\118\002\125\002\126\002\127\002\128\002\
    \130\002\131\002\132\002\133\002\134\002\135\002\136\002\137\002\
    \138\002\139\002\140\002\141\002\143\002\144\002\145\002\146\002\
    \147\002\148\002\149\002\150\002\151\002\152\002\128\002\130\002\
    \153\002\154\002\155\002\154\002\157\002\156\002\160\002\159\002\
    \161\002\159\002\162\002\161\002\128\002\130\002\156\002\156\002\
    \148\002\149\002\158\002\160\002\164\002\158\002\168\002\154\002\
    \159\002\159\002\158\002\162\002\158\002\169\002\163\002\170\002\
    \162\002\163\002\171\002\166\002\172\002\163\002\173\002\165\002\
    \174\002\163\002\175\002\176\002\163\002\166\002\177\002\165\002\
    \178\002\179\002\149\002\149\002\149\002\149\002\149\002\149\002\
    \149\002\149\002\149\002\165\002\180\002\149\002\149\002\167\002\
    \149\002\149\002\149\002\149\002\149\002\149\002\167\002\149\002\
    \181\002\182\002\183\002\184\002\185\002\186\002\187\002\188\002\
    \167\002\189\002\190\002\128\002\130\002\128\002\130\002\167\002\
    \191\002\192\002\193\002\194\002\195\002\196\002\197\002\199\002\
    \197\002\198\002\200\002\201\002\202\002\187\002\188\002\203\002\
    \198\002\204\002\205\002\206\002\207\002\208\002\209\002\210\002\
    \211\002\212\002\213\002\214\002\215\002\216\002\217\002\218\002\
    \219\002\221\002\220\002\223\002\222\002\224\002\225\002\221\002\
    \226\002\227\002\219\002\228\002\229\002\230\002\231\002\232\002\
    \219\002\220\002\222\002\233\002\234\002\235\002\236\002\237\002\
    \238\002\239\002\240\002\241\002\188\002\242\002\243\002\244\002\
    \227\002\245\002\228\002\246\002\247\002\248\002\249\002\250\002\
    \251\002\252\002\253\002\254\002\255\002\000\003\001\003\002\003\
    \003\003\004\003\006\003\002\003\005\003\243\002\244\002\005\003\
    \007\003\008\003\009\003\010\003\011\003\012\003\013\003\014\003\
    \015\003\016\003\007\003\017\003\018\003\019\003\020\003\021\003\
    \022\003\023\003\024\003\025\003\026\003\027\003\228\002\028\003\
    \029\003\030\003\031\003\005\003\032\003\033\003\034\003\035\003\
    \036\003\037\003\038\003\244\002\005\003\039\003\040\003\041\003\
    \005\003\042\003\043\003\044\003\045\003\046\003\128\002\130\002\
    \047\003\048\003\049\003\050\003\051\003\052\003\053\003\054\003\
    \055\003\056\003\057\003\058\003\059\003\061\003\060\003\062\003\
    \063\003\058\003\060\003\064\003\065\003\066\003\067\003\068\003\
    \069\003\062\003\070\003\071\003\072\003\073\003\074\003\075\003\
    \077\003\078\003\079\003\080\003\081\003\082\003\083\003\084\003\
    \085\003\086\003\087\003\088\003\089\003\090\003\091\003\092\003\
    \093\003\094\003\095\003\096\003\097\003\098\003\099\003\100\003\
    \101\003\102\003\103\003\104\003\105\003\103\003\106\003\107\003\
    \108\003\109\003\110\003\111\003\112\003\113\003\114\003\093\003\
    \094\003\115\003\116\003\117\003\118\003\119\003\120\003\121\003\
    \122\003\123\003\124\003\125\003\126\003\127\003\128\003\129\003\
    \130\003\131\003\130\003\132\003\133\003\135\003\136\003\137\003\
    \115\003\116\003\138\003\139\003\140\003\141\003\142\003\143\003\
    \144\003\145\003\146\003\147\003\148\003\149\003\150\003\151\003\
    \152\003\094\003\153\003\154\003\155\003\156\003\157\003\158\003\
    \159\003\160\003\161\003\162\003\163\003\164\003\165\003\166\003\
    \167\003\168\003\169\003\170\003\171\003\155\003\172\003\173\003\
    \174\003\175\003\116\003\176\003\177\003\178\003\116\003\179\003\
    \180\003\181\003\182\003\183\003\184\003\169\003\185\003\186\003\
    \187\003\188\003\189\003\190\003\191\003\192\003\193\003\193\003\
    \194\003\195\003\196\003\197\003\198\003\195\003\199\003\200\003\
    \201\003\202\003\203\003\204\003\205\003\206\003\193\003\207\003\
    \208\003\209\003\210\003\211\003\212\003\213\003\214\003\215\003\
    \216\003\217\003\218\003\219\003\220\003\221\003\222\003\223\003\
    \224\003\225\003\226\003\227\003\228\003\229\003\230\003\231\003\
    \232\003\233\003\234\003\212\003\213\003\233\003\235\003\236\003\
    \237\003\238\003\239\003\240\003\241\003\242\003\243\003\244\003\
    \245\003\246\003\247\003\248\003\249\003\250\003\251\003\252\003\
    \253\003\254\003\255\003\000\004\001\004\002\004\003\004\004\004\
    \005\004\006\004\007\004\008\004\009\004\010\004\011\004\012\004\
    \013\004\014\004\015\004\016\004\017\004\018\004\019\004\020\004\
    \021\004\022\004\023\004\024\004\025\004\026\004\027\004\028\004\
    \213\003\029\004\030\004\031\004\032\004\033\004\034\004\035\004\
    \036\004\015\004\016\004\037\004\038\004\039\004\040\004\041\004\
    \042\004\036\004\043\004\042\004\044\004\045\004\046\004\047\004\
    \048\004\049\004\050\004\051\004\052\004\053\004\054\004\055\004\
    \056\004\057\004\058\004\059\004\060\004\061\004\062\004\062\004\
    \064\004\065\004\062\004\066\004\067\004\068\004\069\004\074\004\
    \068\004\070\004\072\004\070\004\071\004\082\004\069\004\073\004\
    \075\004\016\004\075\004\079\004\075\004\062\004\097\004\064\004\
    \077\004\073\004\069\004\072\004\071\004\062\004\075\004\070\004\
    \072\004\079\004\080\004\075\004\075\004\077\004\078\004\099\004\
    \090\004\078\004\101\004\081\004\100\004\078\004\081\004\100\004\
    \102\004\078\004\080\004\081\004\078\004\081\004\062\004\062\004\
    \062\004\062\004\062\004\062\004\062\004\062\004\062\004\090\004\
    \103\004\062\004\062\004\062\004\062\004\062\004\062\004\062\004\
    \062\004\062\004\062\004\062\004\076\004\083\004\104\004\084\004\
    \087\004\087\004\105\004\076\004\087\004\106\004\107\004\083\004\
    \084\004\084\004\091\004\091\004\083\004\076\004\091\004\095\004\
    \095\004\108\004\109\004\095\004\076\004\098\004\098\004\087\004\
    \098\004\110\004\111\004\112\004\106\004\107\004\113\004\114\004\
    \115\004\091\004\117\004\118\004\119\004\098\004\095\004\120\004\
    \121\004\122\004\123\004\120\004\090\004\124\004\125\004\126\004\
    \128\004\129\004\130\004\131\004\132\004\133\004\134\004\135\004\
    \136\004\137\004\138\004\139\004\140\004\141\004\107\004\142\004\
    \143\004\145\004\146\004\147\004\148\004\149\004\150\004\151\004\
    \152\004\151\004\153\004\154\004\155\004\150\004\156\004\157\004\
    \158\004\138\004\139\004\159\004\087\004\160\004\162\004\161\004\
    \164\004\163\004\165\004\166\004\162\004\167\004\091\004\160\004\
    \168\004\169\004\170\004\160\004\171\004\160\004\161\004\163\004\
    \172\004\173\004\174\004\175\004\176\004\177\004\178\004\179\004\
    \180\004\181\004\182\004\183\004\184\004\185\004\186\004\187\004\
    \188\004\189\004\190\004\191\004\192\004\193\004\194\004\172\004\
    \173\004\195\004\196\004\197\004\198\004\199\004\139\004\199\004\
    \200\004\201\004\202\004\203\004\204\004\205\004\208\004\188\004\
    \189\004\206\004\207\004\209\004\210\004\211\004\212\004\213\004\
    \214\004\215\004\216\004\217\004\218\004\219\004\220\004\221\004\
    \222\004\223\004\224\004\225\004\226\004\062\004\227\004\228\004\
    \206\004\207\004\229\004\218\004\173\004\230\004\231\004\218\004\
    \232\004\233\004\234\004\235\004\236\004\189\004\237\004\238\004\
    \239\004\240\004\241\004\242\004\243\004\244\004\245\004\246\004\
    \247\004\248\004\249\004\250\004\251\004\252\004\253\004\254\004\
    \255\004\000\005\001\005\002\005\003\005\000\005\004\005\005\005\
    \006\005\007\005\008\005\243\004\244\004\009\005\010\005\207\004\
    \011\005\012\005\013\005\014\005\015\005\016\005\017\005\018\005\
    \019\005\020\005\021\005\022\005\023\005\024\005\025\005\026\005\
    \027\005\028\005\029\005\030\005\031\005\032\005\033\005\034\005\
    \035\005\036\005\037\005\038\005\039\005\040\005\041\005\244\004\
    \042\005\244\004\043\005\044\005\045\005\244\004\046\005\047\005\
    \048\005\049\005\050\005\244\004\048\005\051\005\244\004\052\005\
    \244\004\244\004\053\005\052\005\054\005\055\005\056\005\057\005\
    \058\005\059\005\060\005\061\005\062\005\063\005\064\005\066\005\
    \067\005\068\005\069\005\066\005\068\005\070\005\071\005\072\005\
    \073\005\074\005\075\005\076\005\077\005\078\005\079\005\080\005\
    \069\005\081\005\082\005\083\005\075\005\084\005\085\005\086\005\
    \087\005\088\005\089\005\090\005\091\005\092\005\093\005\094\005\
    \068\005\095\005\096\005\097\005\098\005\099\005\100\005\101\005\
    \102\005\068\005\103\005\104\005\105\005\068\005\106\005\107\005\
    \108\005\109\005\110\005\111\005\112\005\113\005\114\005\115\005\
    \116\005\117\005\118\005\119\005\120\005\121\005\122\005\123\005\
    \123\005\124\005\125\005\123\005\126\005\127\005\128\005\129\005\
    \130\005\123\005\131\005\132\005\133\005\134\005\135\005\136\005\
    \137\005\138\005\139\005\140\005\141\005\142\005\143\005\144\005\
    \145\005\146\005\147\005\148\005\149\005\150\005\151\005\152\005\
    \153\005\154\005\141\005\155\005\134\005\135\005\156\005\157\005\
    \158\005\159\005\160\005\161\005\162\005\163\005\164\005\165\005\
    \166\005\167\005\168\005\169\005\170\005\171\005\172\005\173\005\
    \154\005\175\005\155\005\174\005\176\005\171\005\178\005\174\005\
    \179\005\180\005\181\005\182\005\183\005\184\005\185\005\186\005\
    \187\005\188\005\169\005\170\005\189\005\190\005\135\005\191\005\
    \192\005\135\005\135\005\193\005\194\005\195\005\196\005\135\005\
    \135\005\197\005\198\005\199\005\200\005\201\005\202\005\203\005\
    \204\005\205\005\206\005\207\005\208\005\209\005\210\005\211\005\
    \212\005\213\005\214\005\215\005\216\005\155\005\216\005\217\005\
    \218\005\219\005\220\005\221\005\201\005\202\005\222\005\223\005\
    \224\005\225\005\170\005\226\005\227\005\228\005\229\005\216\005\
    \230\005\231\005\232\005\233\005\234\005\235\005\236\005\237\005\
    \219\005\220\005\238\005\239\005\240\005\241\005\242\005\243\005\
    \244\005\245\005\246\005\247\005\248\005\229\005\249\005\230\005\
    \250\005\251\005\252\005\253\005\254\005\255\005\000\006\001\006\
    \002\006\003\006\004\006\202\005\005\006\006\006\007\006\008\006\
    \009\006\010\006\011\006\012\006\013\006\014\006\015\006\016\006\
    \017\006\018\006\220\005\019\006\255\005\000\006\020\006\021\006\
    \230\005\220\005\230\005\022\006\023\006\024\006\014\006\025\006\
    \026\006\027\006\028\006\029\006\030\006\031\006\032\006\033\006\
    \034\006\035\006\036\006\037\006\230\005\038\006\039\006\040\006\
    \041\006\042\006\043\006\044\006\045\006\046\006\000\006\047\006\
    \048\006\049\006\050\006\051\006\052\006\053\006\054\006\055\006\
    \056\006\057\006\058\006\060\006\061\006\062\006\063\006\064\006\
    \065\006\066\006\068\006\067\006\069\006\070\006\071\006\048\006\
    \049\006\067\006\072\006\073\006\069\006\069\006\074\006\075\006\
    \076\006\069\006\077\006\078\006\072\006\079\006\080\006\081\006\
    \082\006\083\006\084\006\085\006\086\006\087\006\088\006\089\006\
    \090\006\091\006\092\006\093\006\094\006\074\006\075\006\095\006\
    \096\006\097\006\098\006\099\006\100\006\101\006\102\006\104\006\
    \105\006\106\006\107\006\074\006\075\006\109\006\108\006\049\006\
    \108\006\110\006\111\006\114\006\113\006\115\006\113\006\116\006\
    \115\006\112\006\110\006\110\006\112\006\102\006\103\006\118\006\
    \114\006\112\006\119\006\112\006\108\006\113\006\113\006\120\006\
    \116\006\117\006\119\006\121\006\117\006\116\006\122\006\123\006\
    \117\006\120\006\121\006\124\006\117\006\103\006\119\006\117\006\
    \125\006\126\006\127\006\128\006\121\006\129\006\130\006\131\006\
    \132\006\133\006\134\006\121\006\135\006\136\006\137\006\138\006\
    \139\006\140\006\143\006\144\006\141\006\145\006\146\006\147\006\
    \148\006\149\006\074\006\075\006\074\006\075\006\103\006\103\006\
    \103\006\103\006\103\006\103\006\103\006\103\006\103\006\142\006\
    \150\006\103\006\103\006\141\006\103\006\103\006\103\006\103\006\
    \103\006\103\006\151\006\103\006\151\006\152\006\153\006\154\006\
    \155\006\156\006\157\006\158\006\152\006\159\006\142\006\160\006\
    \161\006\162\006\163\006\164\006\165\006\166\006\167\006\168\006\
    \169\006\170\006\171\006\172\006\173\006\175\006\174\006\177\006\
    \176\006\178\006\179\006\175\006\180\006\183\006\173\006\181\006\
    \182\006\184\006\185\006\186\006\173\006\174\006\176\006\187\006\
    \188\006\189\006\190\006\191\006\192\006\193\006\194\006\195\006\
    \196\006\197\006\199\006\198\006\142\006\200\006\181\006\182\006\
    \201\006\202\006\203\006\204\006\205\006\206\006\207\006\208\006\
    \209\006\210\006\211\006\212\006\213\006\214\006\216\006\212\006\
    \197\006\215\006\198\006\217\006\215\006\218\006\219\006\220\006\
    \221\006\222\006\223\006\224\006\225\006\217\006\226\006\227\006\
    \228\006\229\006\230\006\231\006\232\006\233\006\234\006\235\006\
    \236\006\237\006\238\006\182\006\239\006\240\006\241\006\242\006\
    \215\006\243\006\244\006\245\006\246\006\074\006\075\006\198\006\
    \247\006\215\006\248\006\249\006\250\006\215\006\251\006\252\006\
    \253\006\254\006\255\006\000\007\001\007\002\007\003\007\004\007\
    \005\007\006\007\007\007\008\007\009\007\010\007\011\007\012\007\
    \013\007\015\007\014\007\016\007\017\007\012\007\014\007\018\007\
    \019\007\020\007\021\007\022\007\023\007\016\007\024\007\025\007\
    \026\007\027\007\028\007\029\007\030\007\031\007\032\007\033\007\
    \034\007\035\007\036\007\037\007\038\007\039\007\040\007\041\007\
    \042\007\043\007\044\007\045\007\046\007\047\007\048\007\049\007\
    \050\007\051\007\052\007\053\007\054\007\055\007\056\007\057\007\
    \058\007\059\007\057\007\060\007\061\007\062\007\063\007\064\007\
    \065\007\066\007\067\007\068\007\047\007\048\007\069\007\070\007\
    \071\007\072\007\073\007\074\007\075\007\076\007\077\007\078\007\
    \079\007\080\007\081\007\082\007\083\007\084\007\085\007\084\007\
    \086\007\087\007\088\007\089\007\090\007\069\007\070\007\091\007\
    \092\007\093\007\094\007\095\007\096\007\097\007\098\007\099\007\
    \100\007\101\007\102\007\103\007\104\007\105\007\048\007\106\007\
    \107\007\108\007\109\007\110\007\111\007\112\007\113\007\114\007\
    \115\007\116\007\117\007\118\007\119\007\120\007\121\007\122\007\
    \123\007\124\007\125\007\109\007\126\007\127\007\128\007\070\007\
    \129\007\130\007\131\007\070\007\132\007\133\007\134\007\135\007\
    \136\007\137\007\138\007\123\007\139\007\140\007\141\007\142\007\
    \143\007\144\007\145\007\146\007\147\007\147\007\148\007\149\007\
    \150\007\151\007\152\007\149\007\153\007\154\007\155\007\156\007\
    \157\007\158\007\159\007\160\007\147\007\161\007\162\007\163\007\
    \164\007\165\007\166\007\167\007\168\007\169\007\170\007\171\007\
    \172\007\173\007\174\007\175\007\176\007\177\007\178\007\179\007\
    \180\007\181\007\182\007\183\007\184\007\185\007\186\007\187\007\
    \188\007\166\007\167\007\187\007\189\007\190\007\191\007\192\007\
    \193\007\194\007\195\007\196\007\197\007\198\007\199\007\200\007\
    \201\007\202\007\203\007\204\007\205\007\206\007\207\007\208\007\
    \209\007\210\007\211\007\212\007\213\007\214\007\215\007\216\007\
    \217\007\218\007\219\007\220\007\221\007\222\007\223\007\224\007\
    \225\007\226\007\227\007\228\007\229\007\230\007\231\007\232\007\
    \234\007\233\007\235\007\227\007\233\007\236\007\167\007\237\007\
    \238\007\239\007\240\007\241\007\242\007\243\007\244\007\245\007\
    \246\007\247\007\248\007\249\007\250\007\251\007\252\007\253\007\
    \254\007\255\007\000\008\001\008\002\008\003\008\004\008\005\008\
    \006\008\007\008\008\008\009\008\010\008\011\008\012\008\013\008\
    \014\008\015\008\016\008\017\008\018\008\019\008\020\008\021\008\
    \022\008\023\008\024\008\025\008\003\008\004\008\026\008\026\008\
    \030\008\031\008\026\008\032\008\032\008\033\008\033\008\032\008\
    \051\008\033\008\034\008\052\008\035\008\035\008\036\008\036\008\
    \035\008\053\008\036\008\037\008\060\008\026\008\062\008\038\008\
    \038\008\044\008\032\008\038\008\033\008\043\008\071\008\026\008\
    \040\008\034\008\026\008\035\008\072\008\036\008\004\008\032\008\
    \041\008\033\008\037\008\034\008\042\008\035\008\038\008\047\008\
    \035\008\026\008\036\008\074\008\037\008\039\008\039\008\040\008\
    \038\008\044\008\050\008\038\008\043\008\043\008\056\008\041\008\
    \073\008\040\008\075\008\042\008\046\008\046\008\047\008\076\008\
    \046\008\041\008\048\008\048\008\039\008\063\008\048\008\055\008\
    \055\008\050\008\066\008\055\008\026\008\056\008\039\008\057\008\
    \078\008\058\008\057\008\046\008\058\008\079\008\050\008\081\008\
    \080\008\048\008\080\008\073\008\063\008\090\008\055\008\088\008\
    \046\008\066\008\088\008\092\008\095\008\095\008\048\008\074\008\
    \095\008\063\008\097\008\055\008\077\008\077\008\066\008\067\008\
    \082\008\079\008\067\008\102\008\075\008\048\008\096\008\096\008\
    \109\008\120\008\076\008\095\008\121\008\127\008\128\008\129\008\
    \130\008\131\008\134\008\095\008\139\008\078\008\113\008\067\008\
    \067\008\067\008\067\008\151\008\067\008\067\008\137\008\067\008\
    \087\008\141\008\154\008\087\008\048\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\155\008\
    \057\008\067\008\077\008\165\008\174\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\192\008\067\008\195\008\067\008\082\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\067\008\067\008\067\008\067\008\067\008\067\008\067\008\
    \067\008\087\008\098\008\098\008\099\008\099\008\098\008\107\008\
    \099\008\196\008\197\008\100\008\100\008\026\008\193\008\100\008\
    \198\008\101\008\101\008\199\008\200\008\101\008\132\008\132\008\
    \182\008\098\008\132\008\099\008\204\008\117\008\107\008\208\008\
    \117\008\098\008\100\008\099\008\182\008\193\008\124\008\044\008\
    \101\008\124\008\100\008\043\008\214\008\132\008\135\008\135\008\
    \101\008\137\008\135\008\218\008\141\008\146\008\146\008\147\008\
    \147\008\146\008\147\008\147\008\161\008\161\008\206\008\161\008\
    \161\008\206\008\232\008\235\008\039\008\135\008\223\008\170\008\
    \170\008\050\008\170\008\170\008\146\008\168\008\168\008\168\008\
    \168\008\168\008\168\008\168\008\168\008\168\008\168\008\168\008\
    \168\008\048\008\124\008\238\008\063\008\223\008\055\008\240\008\
    \241\008\066\008\169\008\169\008\169\008\169\008\169\008\169\008\
    \169\008\169\008\169\008\169\008\183\008\070\008\187\008\212\008\
    \244\008\247\008\250\008\253\008\109\008\212\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \254\008\113\008\000\009\003\009\004\009\067\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\070\008\012\009\070\008\025\009\070\008\031\009\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\070\008\070\008\070\008\070\008\070\008\070\008\070\008\
    \070\008\083\008\202\008\202\008\228\008\228\008\202\008\228\008\
    \228\008\050\009\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\008\009\008\009\054\009\008\009\
    \008\009\202\008\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\083\008\182\008\083\008\
    \058\009\083\008\117\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\083\008\083\008\083\008\
    \083\008\083\008\083\008\083\008\083\008\144\008\144\008\061\009\
    \242\008\144\008\062\009\065\009\124\008\177\008\177\008\177\008\
    \177\008\177\008\177\008\177\008\177\008\177\008\177\008\177\008\
    \177\008\066\009\063\009\067\009\144\008\181\008\181\008\242\008\
    \069\009\205\008\205\008\071\009\144\008\205\008\144\008\178\008\
    \178\008\178\008\178\008\178\008\178\008\178\008\178\008\178\008\
    \178\008\063\009\075\009\077\009\181\008\080\009\070\008\212\008\
    \205\008\082\009\183\008\084\009\187\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\144\008\
    \086\009\144\008\089\009\144\008\242\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\144\008\
    \144\008\144\008\144\008\144\008\144\008\144\008\144\008\166\008\
    \166\008\166\008\166\008\166\008\166\008\166\008\166\008\166\008\
    \166\008\166\008\166\008\166\008\166\008\166\008\166\008\114\009\
    \115\009\126\009\083\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\148\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\148\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\148\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\148\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\148\008\148\008\148\008\148\008\148\008\148\008\
    \148\008\148\008\076\009\136\009\145\009\076\009\144\008\144\008\
    \020\009\020\009\020\009\020\009\020\009\020\009\020\009\020\009\
    \020\009\020\009\144\008\144\008\151\009\149\008\149\008\149\008\
    \149\008\149\008\149\008\149\008\149\008\149\008\149\008\149\008\
    \149\008\149\008\149\008\149\008\149\008\149\008\144\008\149\008\
    \149\008\149\008\149\008\149\008\149\008\149\008\149\008\149\008\
    \149\008\149\008\149\008\149\008\149\008\149\008\149\008\149\008\
    \149\008\149\008\149\008\149\008\149\008\149\008\149\008\149\008\
    \149\008\149\008\149\008\149\008\144\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\181\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\255\255\150\008\150\008\150\008\
    \150\008\150\008\150\008\150\008\150\008\152\008\216\008\216\008\
    \056\009\255\255\216\008\072\009\152\008\112\009\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \021\009\021\009\048\009\021\009\021\009\216\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\087\009\255\255\255\255\087\009\152\008\255\255\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\152\008\152\008\152\008\152\008\152\008\152\008\152\008\
    \152\008\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\255\255\255\255\255\255\255\255\160\008\
    \255\255\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\160\008\160\008\160\008\160\008\160\008\
    \160\008\160\008\160\008\219\008\219\008\027\009\027\009\219\008\
    \027\009\027\009\092\009\122\009\122\009\092\009\122\009\122\009\
    \152\008\152\008\132\009\132\009\255\255\132\009\132\009\245\008\
    \245\008\255\255\219\008\245\008\152\008\152\008\175\008\175\008\
    \175\008\175\008\175\008\175\008\175\008\175\008\175\008\175\008\
    \175\008\175\008\175\008\175\008\175\008\175\008\245\008\255\255\
    \152\008\248\008\248\008\255\255\255\255\248\008\001\009\001\009\
    \005\009\005\009\001\009\118\009\005\009\035\009\035\009\035\009\
    \035\009\035\009\035\009\035\009\035\009\035\009\035\009\255\255\
    \248\008\255\255\160\008\160\008\255\255\001\009\255\255\005\009\
    \056\009\072\009\118\009\255\255\255\255\112\009\160\008\160\008\
    \017\009\017\009\017\009\017\009\017\009\017\009\017\009\017\009\
    \017\009\017\009\017\009\017\009\017\009\017\009\017\009\017\009\
    \048\009\255\255\160\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\162\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\162\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\162\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\162\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\162\008\162\008\162\008\162\008\162\008\162\008\
    \162\008\162\008\163\008\163\008\163\008\163\008\163\008\163\008\
    \163\008\163\008\163\008\163\008\163\008\163\008\163\008\163\008\
    \163\008\163\008\163\008\052\009\163\008\163\008\163\008\163\008\
    \163\008\163\008\163\008\163\008\163\008\163\008\163\008\163\008\
    \163\008\163\008\163\008\163\008\163\008\163\008\163\008\163\008\
    \163\008\163\008\163\008\163\008\163\008\163\008\163\008\163\008\
    \163\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \255\255\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \092\009\164\008\164\008\164\008\164\008\164\008\164\008\164\008\
    \164\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\167\008\167\008\167\008\167\008\167\008\167\008\167\008\
    \167\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \171\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \171\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \171\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \171\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \171\008\171\008\171\008\171\008\171\008\171\008\171\008\171\008\
    \172\008\172\008\172\008\172\008\172\008\172\008\172\008\172\008\
    \172\008\172\008\172\008\172\008\172\008\172\008\172\008\172\008\
    \172\008\052\009\172\008\172\008\172\008\172\008\172\008\172\008\
    \172\008\172\008\172\008\172\008\172\008\172\008\172\008\172\008\
    \172\008\172\008\172\008\172\008\172\008\172\008\172\008\172\008\
    \172\008\172\008\172\008\172\008\172\008\172\008\172\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\255\255\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\255\255\173\008\
    \173\008\173\008\173\008\173\008\173\008\173\008\173\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\176\008\
    \176\008\176\008\176\008\176\008\176\008\176\008\176\008\179\008\
    \179\008\255\255\221\008\179\008\184\008\184\008\164\009\060\009\
    \184\008\185\008\185\008\189\008\189\008\093\009\166\009\189\008\
    \093\009\236\008\236\008\167\009\068\009\236\008\179\008\098\009\
    \179\008\221\008\098\009\184\008\099\009\184\008\060\009\099\009\
    \185\008\255\255\189\008\255\255\179\008\189\008\189\008\189\008\
    \236\008\184\008\189\008\068\009\189\008\224\008\224\008\141\009\
    \141\009\224\008\141\009\141\009\255\255\068\009\236\008\019\009\
    \019\009\019\009\019\009\019\009\019\009\019\009\019\009\019\009\
    \019\009\019\009\019\009\255\255\224\008\255\255\224\008\070\009\
    \070\009\083\009\083\009\070\009\224\008\083\009\255\255\255\255\
    \255\255\255\255\224\008\039\009\039\009\039\009\039\009\039\009\
    \039\009\039\009\039\009\039\009\039\009\189\008\070\009\255\255\
    \083\009\255\255\255\255\060\009\255\255\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\255\255\
    \255\255\224\008\255\255\224\008\255\255\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\224\008\
    \224\008\224\008\224\008\224\008\224\008\224\008\224\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\229\008\229\008\
    \229\008\229\008\229\008\229\008\229\008\229\008\032\009\032\009\
    \032\009\032\009\032\009\032\009\032\009\032\009\032\009\032\009\
    \032\009\032\009\032\009\032\009\032\009\032\009\116\009\255\255\
    \101\009\255\008\255\008\101\009\255\255\255\008\224\008\224\008\
    \043\009\043\009\043\009\043\009\043\009\043\009\043\009\043\009\
    \043\009\043\009\224\008\224\008\119\009\116\009\179\008\128\009\
    \255\008\221\008\064\009\184\008\164\009\255\255\255\255\116\009\
    \185\008\255\255\189\008\093\009\166\009\255\255\224\008\255\255\
    \236\008\167\009\255\255\119\009\255\255\098\009\128\009\255\255\
    \255\255\064\009\099\009\230\008\230\008\230\008\230\008\230\008\
    \230\008\230\008\230\008\230\008\230\008\230\008\230\008\230\008\
    \230\008\230\008\230\008\230\008\224\008\230\008\230\008\230\008\
    \230\008\230\008\230\008\230\008\230\008\230\008\230\008\230\008\
    \230\008\230\008\230\008\230\008\230\008\230\008\230\008\230\008\
    \230\008\230\008\230\008\230\008\230\008\230\008\230\008\230\008\
    \230\008\230\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\064\009\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\255\255\231\008\231\008\231\008\231\008\231\008\231\008\
    \231\008\231\008\233\008\239\008\239\008\255\255\255\255\239\008\
    \255\255\233\008\255\255\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\147\009\147\009\255\255\
    \147\009\147\009\239\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\255\255\255\255\
    \255\255\255\255\233\008\255\255\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\233\008\233\008\
    \233\008\233\008\233\008\233\008\233\008\233\008\101\009\239\008\
    \255\008\243\008\243\008\255\255\255\255\243\008\255\255\010\009\
    \010\009\010\009\010\009\010\009\010\009\010\009\010\009\010\009\
    \010\009\010\009\010\009\010\009\010\009\010\009\010\009\010\009\
    \243\008\010\009\010\009\010\009\010\009\010\009\010\009\010\009\
    \010\009\010\009\010\009\010\009\010\009\010\009\010\009\010\009\
    \010\009\010\009\010\009\010\009\010\009\010\009\010\009\010\009\
    \010\009\010\009\010\009\010\009\010\009\010\009\034\009\034\009\
    \034\009\034\009\034\009\034\009\034\009\034\009\034\009\034\009\
    \034\009\034\009\255\255\255\255\255\255\233\008\233\008\255\255\
    \255\255\044\009\085\009\085\009\255\255\234\008\085\009\255\255\
    \255\255\233\008\233\008\255\255\234\008\243\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \044\009\085\009\044\009\255\255\255\255\233\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\255\255\255\255\255\255\255\255\234\008\255\255\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\234\008\234\008\234\008\234\008\234\008\234\008\234\008\
    \234\008\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\255\255\255\255\104\009\255\255\007\009\
    \104\009\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\007\009\007\009\007\009\007\009\007\009\
    \007\009\007\009\007\009\038\009\038\009\038\009\038\009\038\009\
    \038\009\038\009\038\009\038\009\038\009\038\009\038\009\255\255\
    \234\008\234\008\255\255\255\255\255\255\255\255\255\255\255\255\
    \106\009\255\255\255\255\106\009\234\008\234\008\036\009\036\009\
    \036\009\036\009\036\009\036\009\036\009\036\009\036\009\036\009\
    \036\009\036\009\036\009\036\009\036\009\036\009\117\009\255\255\
    \234\008\040\009\040\009\040\009\040\009\040\009\040\009\040\009\
    \040\009\040\009\040\009\040\009\040\009\040\009\040\009\040\009\
    \040\009\255\255\255\255\255\255\255\255\117\009\255\255\255\255\
    \255\255\255\255\007\009\007\009\255\255\255\255\255\255\117\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\007\009\007\009\
    \042\009\042\009\042\009\042\009\042\009\042\009\042\009\042\009\
    \042\009\042\009\042\009\042\009\255\255\255\255\255\255\255\255\
    \044\009\255\255\007\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\009\
    \009\009\009\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\255\255\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\255\255\011\009\011\009\011\009\011\009\011\009\011\009\
    \011\009\011\009\013\009\104\009\255\255\255\255\255\255\255\255\
    \255\255\013\009\255\255\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\106\009\255\255\
    \255\255\255\255\013\009\255\255\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\013\009\013\009\
    \013\009\013\009\013\009\013\009\013\009\013\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\018\009\018\009\
    \018\009\018\009\018\009\018\009\018\009\018\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\013\009\013\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\014\009\255\255\255\255\
    \255\255\013\009\013\009\255\255\014\009\255\255\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \255\255\255\255\255\255\255\255\255\255\013\009\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\107\009\255\255\255\255\107\009\014\009\255\255\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\014\009\014\009\014\009\014\009\014\009\014\009\014\009\
    \014\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \022\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \022\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \022\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \022\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \022\009\022\009\022\009\022\009\022\009\022\009\022\009\022\009\
    \078\009\078\009\102\009\094\009\078\009\102\009\094\009\137\009\
    \137\009\137\009\137\009\137\009\137\009\137\009\137\009\137\009\
    \137\009\137\009\137\009\137\009\137\009\137\009\137\009\078\009\
    \014\009\014\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \015\009\255\255\255\255\255\255\014\009\014\009\094\009\015\009\
    \255\255\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\255\255\255\255\255\255\255\255\255\255\
    \014\009\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\078\009\102\009\255\255\255\255\
    \015\009\255\255\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\015\009\015\009\015\009\015\009\
    \015\009\015\009\015\009\015\009\023\009\023\009\023\009\023\009\
    \023\009\023\009\023\009\023\009\023\009\023\009\023\009\023\009\
    \023\009\023\009\023\009\023\009\023\009\255\255\023\009\023\009\
    \023\009\023\009\023\009\023\009\023\009\023\009\023\009\023\009\
    \023\009\023\009\023\009\023\009\023\009\023\009\023\009\023\009\
    \023\009\023\009\023\009\023\009\023\009\023\009\023\009\023\009\
    \023\009\023\009\023\009\255\255\255\255\110\009\095\009\107\009\
    \110\009\095\009\139\009\139\009\139\009\139\009\139\009\139\009\
    \139\009\139\009\139\009\139\009\139\009\139\009\255\255\255\255\
    \255\255\255\255\255\255\015\009\015\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\016\009\255\255\255\255\255\255\015\009\
    \015\009\095\009\016\009\255\255\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\255\255\255\255\
    \255\255\255\255\255\255\015\009\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\078\009\
    \102\009\094\009\255\255\016\009\255\255\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\016\009\
    \016\009\016\009\016\009\016\009\016\009\016\009\016\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\255\255\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\255\255\024\009\
    \024\009\024\009\024\009\024\009\024\009\024\009\024\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\016\009\016\009\
    \111\009\255\255\255\255\111\009\255\255\255\255\026\009\255\255\
    \255\255\255\255\016\009\016\009\255\255\026\009\255\255\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\255\255\255\255\255\255\255\255\255\255\016\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\255\255\110\009\095\009\255\255\026\009\255\255\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\026\009\026\009\026\009\026\009\026\009\026\009\
    \026\009\026\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\028\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\028\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\028\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\028\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\028\009\028\009\028\009\028\009\028\009\028\009\028\009\
    \028\009\140\009\140\009\140\009\140\009\140\009\140\009\140\009\
    \140\009\140\009\140\009\154\009\154\009\154\009\154\009\154\009\
    \154\009\154\009\154\009\154\009\154\009\154\009\154\009\255\255\
    \255\255\026\009\026\009\155\009\155\009\155\009\155\009\155\009\
    \155\009\155\009\155\009\155\009\155\009\026\009\026\009\255\255\
    \029\009\029\009\029\009\029\009\029\009\029\009\029\009\029\009\
    \029\009\029\009\029\009\029\009\029\009\029\009\029\009\029\009\
    \029\009\026\009\029\009\029\009\029\009\029\009\029\009\029\009\
    \029\009\029\009\029\009\029\009\029\009\029\009\029\009\029\009\
    \029\009\029\009\029\009\029\009\029\009\029\009\029\009\029\009\
    \029\009\029\009\029\009\029\009\029\009\029\009\029\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\111\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\255\255\030\009\
    \030\009\030\009\030\009\030\009\030\009\030\009\030\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\033\009\
    \033\009\033\009\033\009\033\009\033\009\033\009\033\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\037\009\
    \037\009\037\009\037\009\037\009\037\009\037\009\037\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\041\009\
    \041\009\041\009\041\009\041\009\041\009\041\009\041\009\090\009\
    \090\009\096\009\129\009\090\009\096\009\100\009\100\009\105\009\
    \105\009\100\009\109\009\105\009\255\255\109\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\090\009\255\255\
    \255\255\129\009\255\255\255\255\100\009\255\255\105\009\255\255\
    \255\255\255\255\255\255\090\009\096\009\158\009\158\009\158\009\
    \158\009\158\009\158\009\158\009\158\009\158\009\158\009\158\009\
    \158\009\255\255\255\255\129\009\159\009\159\009\159\009\159\009\
    \159\009\159\009\159\009\159\009\159\009\159\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \255\255\100\009\255\255\105\009\120\009\109\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \120\009\120\009\120\009\120\009\120\009\120\009\120\009\120\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\123\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\123\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\123\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\123\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\123\009\
    \123\009\123\009\123\009\123\009\123\009\123\009\123\009\152\009\
    \152\009\152\009\152\009\152\009\152\009\152\009\152\009\152\009\
    \152\009\152\009\152\009\152\009\152\009\152\009\152\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\120\009\
    \120\009\163\009\163\009\163\009\163\009\163\009\163\009\163\009\
    \163\009\163\009\163\009\120\009\120\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\156\009\156\009\156\009\
    \156\009\156\009\156\009\156\009\156\009\255\255\255\255\120\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \160\009\160\009\160\009\160\009\160\009\160\009\160\009\160\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\090\009\096\009\
    \255\255\255\255\255\255\255\255\100\009\120\009\105\009\255\255\
    \109\009\124\009\124\009\124\009\124\009\124\009\124\009\124\009\
    \124\009\124\009\124\009\124\009\124\009\124\009\124\009\124\009\
    \124\009\124\009\255\255\124\009\124\009\124\009\124\009\124\009\
    \124\009\124\009\124\009\124\009\124\009\124\009\124\009\124\009\
    \124\009\124\009\124\009\124\009\124\009\124\009\124\009\124\009\
    \124\009\124\009\124\009\124\009\124\009\124\009\124\009\124\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\255\255\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\127\009\
    \125\009\125\009\125\009\125\009\125\009\125\009\125\009\125\009\
    \162\009\162\009\162\009\162\009\162\009\162\009\162\009\162\009\
    \162\009\162\009\162\009\162\009\255\255\255\255\127\009\255\255\
    \255\255\255\255\255\255\255\255\255\255\127\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\127\009\255\255\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\255\255\255\255\255\255\255\255\255\255\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\255\255\255\255\255\255\255\255\127\009\255\255\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\127\009\127\009\127\009\127\009\127\009\127\009\127\009\
    \127\009\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\255\255\255\255\255\255\255\255\130\009\
    \255\255\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\130\009\130\009\130\009\130\009\130\009\
    \130\009\130\009\130\009\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \127\009\127\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\127\009\127\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \127\009\255\255\255\255\255\255\255\255\131\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\130\009\130\009\131\009\255\255\255\255\255\255\
    \255\255\255\255\255\255\131\009\255\255\255\255\130\009\130\009\
    \255\255\255\255\131\009\255\255\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\255\255\
    \255\255\255\255\130\009\255\255\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\255\255\
    \255\255\255\255\255\255\131\009\255\255\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\131\009\
    \131\009\131\009\131\009\131\009\131\009\131\009\131\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\133\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\133\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\133\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\133\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\133\009\133\009\
    \133\009\133\009\133\009\133\009\133\009\133\009\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\131\009\131\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\131\009\131\009\255\255\134\009\134\009\134\009\
    \134\009\134\009\134\009\134\009\134\009\134\009\134\009\134\009\
    \134\009\134\009\134\009\134\009\134\009\134\009\131\009\134\009\
    \134\009\134\009\134\009\134\009\134\009\134\009\134\009\134\009\
    \134\009\134\009\134\009\134\009\134\009\134\009\134\009\134\009\
    \134\009\134\009\134\009\134\009\134\009\134\009\134\009\134\009\
    \134\009\134\009\134\009\134\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\255\255\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\255\255\135\009\135\009\135\009\135\009\
    \135\009\135\009\135\009\135\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\138\009\138\009\138\009\138\009\
    \138\009\138\009\138\009\138\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\142\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\142\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\142\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\142\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\142\009\142\009\142\009\142\009\142\009\
    \142\009\142\009\142\009\143\009\143\009\143\009\143\009\143\009\
    \143\009\143\009\143\009\143\009\143\009\143\009\143\009\143\009\
    \143\009\143\009\143\009\143\009\255\255\143\009\143\009\143\009\
    \143\009\143\009\143\009\143\009\143\009\143\009\143\009\143\009\
    \143\009\143\009\143\009\143\009\143\009\143\009\143\009\143\009\
    \143\009\143\009\143\009\143\009\143\009\143\009\143\009\143\009\
    \143\009\143\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\255\255\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\146\009\144\009\144\009\144\009\144\009\144\009\144\009\
    \144\009\144\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \146\009\255\255\255\255\255\255\255\255\255\255\255\255\146\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\146\009\255\255\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\255\255\255\255\255\255\255\255\255\255\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\255\255\255\255\255\255\255\255\146\009\
    \255\255\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\146\009\146\009\146\009\146\009\146\009\
    \146\009\146\009\146\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\148\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\148\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\148\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\148\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\148\009\148\009\148\009\148\009\148\009\148\009\
    \148\009\148\009\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\146\009\146\009\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\146\009\146\009\
    \255\255\149\009\149\009\149\009\149\009\149\009\149\009\149\009\
    \149\009\149\009\149\009\149\009\149\009\149\009\149\009\149\009\
    \149\009\149\009\146\009\149\009\149\009\149\009\149\009\149\009\
    \149\009\149\009\149\009\149\009\149\009\149\009\149\009\149\009\
    \149\009\149\009\149\009\149\009\149\009\149\009\149\009\149\009\
    \149\009\149\009\149\009\149\009\149\009\149\009\149\009\149\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\255\255\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\255\255\
    \150\009\150\009\150\009\150\009\150\009\150\009\150\009\150\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \153\009\153\009\153\009\153\009\153\009\153\009\153\009\153\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \157\009\157\009\157\009\157\009\157\009\157\009\157\009\157\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \161\009\161\009\161\009\161\009\161\009\161\009\161\009\161\009\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_base_code = 
   "";
  Lexing.lex_backtrk_code = 
   "";
  Lexing.lex_default_code = 
   "";
  Lexing.lex_trans_code = 
   "";
  Lexing.lex_check_code = 
   "";
  Lexing.lex_code = 
   "";
}

let rec coq_bol lexbuf =
    __ocaml_lex_coq_bol_rec lexbuf 0
and __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 461 "tools/coqdoc/cpretty.mll"
      ( if not (!in_proof <> None && (!Cdglobals.gallina || !Cdglobals.light))
        then Output.empty_line_of_code ();
        coq_bol lexbuf )
# 3209 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 465 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3217 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 470 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc (); comments lexbuf; Output.end_doc ();
	Output.start_coq (); coq lexbuf )
# 3223 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 473 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf; coq_bol lexbuf )
# 3228 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 475 "tools/coqdoc/cpretty.mll"
      ( begin_show (); coq_bol lexbuf )
# 3233 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 477 "tools/coqdoc/cpretty.mll"
      ( end_show (); coq_bol lexbuf )
# 3238 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 479 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	coq_bol lexbuf )
# 3247 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 485 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    let eol = skip_to_dot lexbuf in
	      if eol then (coq_bol lexbuf) else coq lexbuf
	  else
	    begin
	      output_indented_keyword s lexbuf;
	      let eol = body lexbuf in
	      if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3261 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 496 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
        output_indented_keyword s lexbuf;
        let eol = body lexbuf in
	in_proof := Some eol;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3270 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 502 "tools/coqdoc/cpretty.mll"
      ( in_proof := Some true;
	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    let s = lexeme lexbuf in
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3283 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 511 "tools/coqdoc/cpretty.mll"
                  (
      let eol =
	if not (!in_proof <> None && !Cdglobals.gallina) then
	  begin backtrack lexbuf; body_bol lexbuf end
	else skip_to_dot lexbuf
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3295 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 520 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3305 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 527 "tools/coqdoc/cpretty.mll"
      (
	in_proof := None;
	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= body lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3315 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 534 "tools/coqdoc/cpretty.mll"
      (	let s = lexeme lexbuf in
	output_indented_keyword s lexbuf;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3323 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 540 "tools/coqdoc/cpretty.mll"
      ( let tok = lexeme lexbuf in
	let s = printing_token_body lexbuf in
	  add_printing_token tok s;
	  coq_bol lexbuf )
# 3331 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 545 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3340 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 552 "tools/coqdoc/cpretty.mll"
      ( remove_printing_token (lexeme lexbuf);
	coq_bol lexbuf )
# 3346 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 555 "tools/coqdoc/cpretty.mll"
      ( eprintf "warning: bad 'remove printing' command at character %d\n"
	  (lexeme_start lexbuf); flush stderr;
	comment_level := 1;
	ignore (comment lexbuf);
	coq_bol lexbuf )
# 3355 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 561 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
	if !Cdglobals.parse_comments then begin
	  let s = lexeme lexbuf in
	  let nbsp,isp = count_spaces s in
	    Output.indentation nbsp;
	    Output.start_comment ();
	end;
	let eol = comment lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3368 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 571 "tools/coqdoc/cpretty.mll"
      ( () )
# 3373 "tools/coqdoc/cpretty.ml"

  | 20 ->
# 573 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body_bol lexbuf end
	  else
	    skip_to_dot lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3384 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_coq_bol_rec lexbuf __ocaml_lex_state

and coq lexbuf =
    __ocaml_lex_coq_rec lexbuf 1086
and __ocaml_lex_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 585 "tools/coqdoc/cpretty.mll"
      ( if not (!in_proof <> None && !Cdglobals.gallina) then Output.line_break(); coq_bol lexbuf )
# 3396 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 587 "tools/coqdoc/cpretty.mll"
      ( Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3404 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 592 "tools/coqdoc/cpretty.mll"
      ( comment_level := 1;
	if !Cdglobals.parse_comments then begin
	  let s = lexeme lexbuf in
	  let nbsp,isp = count_spaces s in
	    Output.indentation nbsp;
	    Output.start_comment ();
	end;
	let eol = comment lexbuf in
	  if eol then coq_bol lexbuf
          else coq lexbuf
      )
# 3419 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 604 "tools/coqdoc/cpretty.mll"
      ( if not !formatted then
	begin
	  (* Isn't this an anomaly *)
	  let s = lexeme lexbuf in
	  let nlsp,s = remove_newline s in
	  let nbsp,isp = count_spaces s in
	  Output.indentation nbsp;
	  let loc = lexeme_start lexbuf + isp + nlsp in
	  Output.sublexer ']' loc;
	  Output.sublexer ']' (loc+1);
	  coq lexbuf
	end )
# 3435 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 617 "tools/coqdoc/cpretty.mll"
      ( () )
# 3440 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 619 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  if !Cdglobals.light && section_or_end s then
	    begin
	      let eol = skip_to_dot lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end
	  else
	    begin
	      Output.ident s None;
	      let eol=body lexbuf in
		if eol then coq_bol lexbuf else coq lexbuf
	    end )
# 3456 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 632 "tools/coqdoc/cpretty.mll"
      ( let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    let s = lexeme lexbuf in
	    let eol =
	      if s.[String.length s - 1] = '.' then false
	      else skip_to_dot lexbuf
	    in
	      eol
	in if eol then coq_bol lexbuf else coq lexbuf )
# 3471 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 643 "tools/coqdoc/cpretty.mll"
           (
      let eol =
	if not !Cdglobals.gallina then
	  begin backtrack lexbuf; body lexbuf end
	else
	  let eol = skip_to_dot lexbuf in
	    if !in_proof <> Some true && eol then
	      Output.line_break ();
	    eol
      in
	in_proof := None;
	if eol then coq_bol lexbuf else coq lexbuf )
# 3487 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 656 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3495 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 661 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	Output.ident s None;
	let eol= start_notation_string lexbuf in
	if eol then coq_bol lexbuf else coq lexbuf )
# 3503 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 666 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	  Output.ident s None;
	let eol = body lexbuf in
	  if eol then coq_bol lexbuf else coq lexbuf )
# 3511 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 670 "tools/coqdoc/cpretty.mll"
           ( Output.char ' '; coq lexbuf )
# 3516 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 672 "tools/coqdoc/cpretty.mll"
      ( () )
# 3521 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 673 "tools/coqdoc/cpretty.mll"
      (	let eol =
	  if not !Cdglobals.gallina then
	    begin backtrack lexbuf; body lexbuf end
	  else
	    skip_to_dot lexbuf
	in
	  if eol then coq_bol lexbuf else coq lexbuf)
# 3532 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_coq_rec lexbuf __ocaml_lex_state

and doc_bol lexbuf =
    __ocaml_lex_doc_bol_rec lexbuf 2074
and __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 685 "tools/coqdoc/cpretty.mll"
      ( let eol, lex = strip_eol (lexeme lexbuf) in
        let lev, s = sec_title lex in
          if (!Cdglobals.lib_subtitles) &&
             (subtitle (Output.get_module false) s) then
            ()
          else
            Output.section lev (fun () -> ignore (doc None (from_string s)));
	    if eol then doc_bol lexbuf else doc None lexbuf )
# 3551 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 694 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
        let lines = (List.length bufs) - 1 in
        let line = 
          match bufs with
          | [] -> eprintf "Internal error bad_split1 - please report\n";
                  exit 1
          | _ -> List.nth bufs lines
        in
          match check_start_list line with
          | Neither -> backtrack_past_newline lexbuf; doc None lexbuf
          | List n -> Output.paragraph ();
                      Output.item 1; doc (Some [n]) lexbuf
          | Rule -> Output.rule (); doc None lexbuf
      )
# 3570 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 710 "tools/coqdoc/cpretty.mll"
      ( Output.paragraph (); doc_bol lexbuf )
# 3575 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 712 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false; verbatim false lexbuf; doc_bol lexbuf )
# 3580 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 714 "tools/coqdoc/cpretty.mll"
      ( true )
# 3585 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 716 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else start_emph ();
        doc None lexbuf )
# 3591 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 719 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; doc None lexbuf )
# 3596 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_doc_bol_rec lexbuf __ocaml_lex_state

and doc_list_bol indents lexbuf =
    __ocaml_lex_doc_list_bol_rec indents lexbuf 2096
and __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 724 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> backtrack lexbuf; doc_bol lexbuf
        | StartLevel n -> Output.item n; doc (Some (take n indents)) lexbuf
        | InLevel (n,true) ->
            let items = List.length indents in
              Output.item (items+1);
              doc (Some (List.append indents [n_spaces])) lexbuf
        | InLevel (_,false) ->
            backtrack lexbuf; doc_bol lexbuf
      )
# 3618 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 736 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim false;
        verbatim false lexbuf;
        doc_list_bol indents lexbuf )
# 3625 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 740 "tools/coqdoc/cpretty.mll"
      ( formatted := true;
        Output.start_inline_coq_block ();
        ignore(body_bol lexbuf);
        Output.end_inline_coq_block ();
        formatted := false;
        doc_list_bol indents lexbuf )
# 3635 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 747 "tools/coqdoc/cpretty.mll"
      ( inf_rules (Some indents) lexbuf )
# 3640 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 749 "tools/coqdoc/cpretty.mll"
      ( (* Like in the doc_bol production, these two productions
           exist only to deal properly with whitespace *)
        Output.paragraph ();
        backtrack_past_newline lexbuf;
        doc_list_bol indents lexbuf )
# 3649 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 755 "tools/coqdoc/cpretty.mll"
      ( let buf' = lexeme lexbuf in
        let buf =
          let bufs = Str.split_delim (Str.regexp "['\n']") buf' in
            match bufs with
              | (_ :: s :: []) -> s
              | (_ :: _ :: s :: _) -> s
              | _ -> eprintf "Internal error bad_split2 - please report\n";
                     exit 1
        in
        let (n_spaces,_) = count_spaces buf in
        match find_level indents n_spaces with
        | InLevel _ ->
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_list_bol indents lexbuf
        | StartLevel n ->
            if n = 1 then
              begin
                Output.stop_item ();
                backtrack_past_newline lexbuf;
                doc_bol lexbuf
              end
            else
              begin
                Output.paragraph ();
                backtrack_past_newline lexbuf;
                doc_list_bol indents lexbuf
              end
        | Before -> 
        (* Here we were at the beginning of a line, and it was blank.
           The next line started before any list items.  So: insert
           a paragraph for the empty line, rewind to whatever's just
           after the newline, then toss over to doc_bol for whatever
           comes next. *)
            Output.stop_item ();
            Output.paragraph ();
            backtrack_past_newline lexbuf;
            doc_bol lexbuf

      )
# 3693 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 796 "tools/coqdoc/cpretty.mll"
      ( let (n_spaces,_) = count_spaces (lexeme lexbuf) in
        match find_level indents n_spaces with
        | Before -> Output.stop_item (); backtrack lexbuf;
                    doc_bol lexbuf
        | StartLevel n ->
            (if n = 1 then
               Output.stop_item ()
             else
               Output.reach_item_level (n-1));
            backtrack lexbuf;
            doc (Some (take (n-1) indents)) lexbuf
        | InLevel (n,_) ->
            Output.reach_item_level n;
            backtrack lexbuf;
            doc (Some (take n indents)) lexbuf
      )
# 3713 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_doc_list_bol_rec indents lexbuf __ocaml_lex_state

and doc indents lexbuf =
    __ocaml_lex_doc_rec indents lexbuf 2115
and __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 816 "tools/coqdoc/cpretty.mll"
      ( Output.char '\n';
        match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 3728 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 821 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments
        then (Output.char '['; Output.char '['; doc indents lexbuf)
        else (formatted := true;
              Output.start_inline_coq_block ();
	      let eol = body_bol lexbuf in
	        Output.end_inline_coq_block (); formatted := false;
	        if eol then
		  match indents with
		  | Some ls -> doc_list_bol ls lexbuf
		  | None -> doc_bol lexbuf
		else doc indents lexbuf))
# 3743 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 833 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 3748 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 835 "tools/coqdoc/cpretty.mll"
      ( Output.proofbox (); doc indents lexbuf )
# 3753 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 836 "tools/coqdoc/cpretty.mll"
         ( url lexbuf; doc indents lexbuf )
# 3758 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 838 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1;  Output.start_inline_coq (); escaped_coq lexbuf;
              Output.end_inline_coq ()); doc indents lexbuf )
# 3765 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 842 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf ;
        let bol_parse = match indents with
                        | Some is -> doc_list_bol is
                        | None   -> doc_bol
        in
        let eol = comment lexbuf in
          if eol then bol_parse lexbuf else doc indents lexbuf
      )
# 3777 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 851 "tools/coqdoc/cpretty.mll"
      ((match indents with
        | Some _ -> Output.stop_item ()
        | None -> ());
       (* this says - if there is a blank line between the two comments,
          insert one in the output too *)
       let lines = List.length (Str.split_delim (Str.regexp "['\n']") 
                                                (lexeme lexbuf))
       in 
         if lines > 2 then Output.paragraph ();
       doc_bol lexbuf
      )
# 3792 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 863 "tools/coqdoc/cpretty.mll"
      ( true )
# 3797 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 865 "tools/coqdoc/cpretty.mll"
      ( false )
# 3802 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 867 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$'
        else (Output.start_latex_math (); escaped_math_latex lexbuf);
        doc indents lexbuf )
# 3809 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 871 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '$';
        Output.char '$'; doc indents lexbuf )
# 3815 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 874 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%'
        else escaped_latex lexbuf; doc indents lexbuf )
# 3821 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 877 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '%';
        Output.char '%'; doc indents lexbuf )
# 3827 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 880 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#'
        else escaped_html lexbuf; doc indents lexbuf )
# 3833 "tools/coqdoc/cpretty.ml"

  | 15 ->
# 883 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '#';
        Output.char '#'; doc indents lexbuf )
# 3839 "tools/coqdoc/cpretty.ml"

  | 16 ->
# 886 "tools/coqdoc/cpretty.mll"
      ( List.iter (fun x -> Output.char (lexeme_char lexbuf x)) [0;1;2];
        doc indents lexbuf)
# 3845 "tools/coqdoc/cpretty.ml"

  | 17 ->
# 889 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0);
        if !Cdglobals.plain_comments then Output.char '_' else  start_emph () ;
        doc indents lexbuf )
# 3852 "tools/coqdoc/cpretty.ml"

  | 18 ->
# 893 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments then Output.char '_' else stop_emph () ;
        Output.char (lexeme_char lexbuf 1);
        doc indents lexbuf )
# 3859 "tools/coqdoc/cpretty.ml"

  | 19 ->
# 897 "tools/coqdoc/cpretty.mll"
      ( Output.start_verbatim true; verbatim true lexbuf; doc_bol lexbuf )
# 3864 "tools/coqdoc/cpretty.ml"

  | 20 ->
# 899 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.plain_comments 
	then Output.char '"' 
	else if in_quote ()
	then stop_quote () 
	else start_quote ();
        doc indents lexbuf )
# 3874 "tools/coqdoc/cpretty.ml"

  | 21 ->
# 906 "tools/coqdoc/cpretty.mll"
      ( false )
# 3879 "tools/coqdoc/cpretty.ml"

  | 22 ->
# 908 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); doc indents lexbuf )
# 3884 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_doc_rec indents lexbuf __ocaml_lex_state

and escaped_math_latex lexbuf =
    __ocaml_lex_escaped_math_latex_rec lexbuf 2157
and __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 913 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 3896 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 914 "tools/coqdoc/cpretty.mll"
        ( Output.stop_latex_math () )
# 3901 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 915 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_math_latex lexbuf )
# 3906 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_escaped_math_latex_rec lexbuf __ocaml_lex_state

and escaped_latex lexbuf =
    __ocaml_lex_escaped_latex_rec lexbuf 2161
and __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 918 "tools/coqdoc/cpretty.mll"
        ( () )
# 3918 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 919 "tools/coqdoc/cpretty.mll"
        ( () )
# 3923 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 920 "tools/coqdoc/cpretty.mll"
        ( Output.latex_char (lexeme_char lexbuf 0); escaped_latex lexbuf )
# 3928 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_escaped_latex_rec lexbuf __ocaml_lex_state

and escaped_html lexbuf =
    __ocaml_lex_escaped_html_rec lexbuf 2165
and __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 923 "tools/coqdoc/cpretty.mll"
        ( () )
# 3940 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 925 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '&'; Output.html_char '#'; escaped_html lexbuf )
# 3945 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 927 "tools/coqdoc/cpretty.mll"
        ( Output.html_char '#'; escaped_html lexbuf )
# 3950 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 928 "tools/coqdoc/cpretty.mll"
        ( () )
# 3955 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 929 "tools/coqdoc/cpretty.mll"
        ( Output.html_char (lexeme_char lexbuf 0); escaped_html lexbuf )
# 3960 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_escaped_html_rec lexbuf __ocaml_lex_state

and verbatim inline lexbuf =
    __ocaml_lex_verbatim_rec inline lexbuf 2172
and __ocaml_lex_verbatim_rec inline lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 932 "tools/coqdoc/cpretty.mll"
                      ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 3972 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 933 "tools/coqdoc/cpretty.mll"
            ( Output.verbatim_char inline '\n'; Output.stop_verbatim inline )
# 3977 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 934 "tools/coqdoc/cpretty.mll"
         ( Output.stop_verbatim inline )
# 3982 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 935 "tools/coqdoc/cpretty.mll"
        ( Output.stop_verbatim inline )
# 3987 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 936 "tools/coqdoc/cpretty.mll"
      ( Output.verbatim_char inline (lexeme_char lexbuf 0); verbatim inline lexbuf )
# 3992 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_verbatim_rec inline lexbuf __ocaml_lex_state

and url lexbuf =
    __ocaml_lex_url_rec lexbuf 2185
and __ocaml_lex_url_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 939 "tools/coqdoc/cpretty.mll"
         ( Output.url (Buffer.contents url_buffer) None; Buffer.clear url_buffer )
# 4004 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 940 "tools/coqdoc/cpretty.mll"
        ( url_name lexbuf )
# 4009 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 941 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_buffer (lexeme_char lexbuf 0); url lexbuf )
# 4014 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_url_rec lexbuf __ocaml_lex_state

and url_name lexbuf =
    __ocaml_lex_url_name_rec lexbuf 2189
and __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 944 "tools/coqdoc/cpretty.mll"
        ( Output.url (Buffer.contents url_buffer) (Some (Buffer.contents url_name_buffer));
	  Buffer.clear url_buffer; Buffer.clear url_name_buffer )
# 4027 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 946 "tools/coqdoc/cpretty.mll"
      ( Buffer.add_char url_name_buffer (lexeme_char lexbuf 0); url_name lexbuf )
# 4032 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_url_name_rec lexbuf __ocaml_lex_state

and escaped_coq lexbuf =
    __ocaml_lex_escaped_coq_rec lexbuf 2192
and __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 952 "tools/coqdoc/cpretty.mll"
      ( decr brackets;
	if !brackets > 0 then
	  (Output.sublexer_in_doc ']'; escaped_coq lexbuf)
	else Tokens.flush_sublexer () )
# 4047 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 957 "tools/coqdoc/cpretty.mll"
      ( incr brackets;
        Output.sublexer_in_doc '['; escaped_coq lexbuf )
# 4053 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 960 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer (); comment_level := 1;
        ignore (comment lexbuf); escaped_coq lexbuf )
# 4059 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 963 "tools/coqdoc/cpretty.mll"
      ( (* likely to be a syntax error: we escape *) backtrack lexbuf )
# 4064 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 965 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer () )
# 4069 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 967 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
        escaped_coq lexbuf )
# 4076 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 971 "tools/coqdoc/cpretty.mll"
      ( let str = lexeme lexbuf in
          Tokens.flush_sublexer();
          (if !Cdglobals.inline_notmono then () 
                                        else Output.end_inline_coq ()); 
          String.iter Output.char str; 
          (if !Cdglobals.inline_notmono then () 
                                        else Output.start_inline_coq ());
          escaped_coq lexbuf )
# 4088 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 980 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer_in_doc (lexeme_char lexbuf 0);
        escaped_coq lexbuf )
# 4094 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_escaped_coq_rec lexbuf __ocaml_lex_state

and comments lexbuf =
    __ocaml_lex_comments_rec lexbuf 2227
and __ocaml_lex_comments_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 987 "tools/coqdoc/cpretty.mll"
      ( Output.char ' '; comments lexbuf )
# 4106 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 989 "tools/coqdoc/cpretty.mll"
      ( let s = lexeme lexbuf in
	let s = String.sub s 1 (String.length s - 2) in
	ignore (doc None (from_string s)); comments lexbuf )
# 4113 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 993 "tools/coqdoc/cpretty.mll"
      ( escaped_coq (from_string (lexeme lexbuf)); comments lexbuf )
# 4118 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 995 "tools/coqdoc/cpretty.mll"
      ( () )
# 4123 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 997 "tools/coqdoc/cpretty.mll"
      ( () )
# 4128 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 999 "tools/coqdoc/cpretty.mll"
      ( Output.char (lexeme_char lexbuf 0); comments lexbuf )
# 4133 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comments_rec lexbuf __ocaml_lex_state

and comment lexbuf =
    __ocaml_lex_comment_rec lexbuf 2237
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1004 "tools/coqdoc/cpretty.mll"
         ( incr comment_level;
	   if !Cdglobals.parse_comments then Output.start_comment ();
	   comment lexbuf )
# 4147 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1007 "tools/coqdoc/cpretty.mll"
                   (
      if !Cdglobals.parse_comments then
	(Output.end_comment (); Output.line_break ());
      decr comment_level; if !comment_level > 0 then comment lexbuf else true )
# 4155 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1011 "tools/coqdoc/cpretty.mll"
         (
      if !Cdglobals.parse_comments then (Output.end_comment ());
      decr comment_level; if !comment_level > 0 then comment lexbuf else false )
# 4162 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1014 "tools/coqdoc/cpretty.mll"
        (
      if !Cdglobals.parse_comments then
	if !Cdglobals.plain_comments then Output.char '['
        else (brackets := 1; Output.start_inline_coq ();
              escaped_coq lexbuf; Output.end_inline_coq ());
      comment lexbuf )
# 4172 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1020 "tools/coqdoc/cpretty.mll"
            (
      if !Cdglobals.parse_comments then
        if !Cdglobals.plain_comments then (Output.char '['; Output.char '[')
        else (formatted := true;
              Output.start_inline_coq_block ();
              let _ = body_bol lexbuf in
                Output.end_inline_coq_block (); formatted := false);
      comment lexbuf)
# 4184 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1029 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments then
          if !Cdglobals.plain_comments then Output.char '$'
          else (Output.start_latex_math (); escaped_math_latex lexbuf);
        comment lexbuf )
# 4192 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1034 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments
        then
          (if !Cdglobals.plain_comments then Output.char '$'; Output.char '$');
        doc None lexbuf )
# 4200 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1039 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments
          then
            if !Cdglobals.plain_comments then Output.char '%'
            else escaped_latex lexbuf; comment lexbuf )
# 4208 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1044 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments
        then
          (if !Cdglobals.plain_comments then Output.char '%'; Output.char '%');
        comment lexbuf )
# 4216 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1049 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments
        then
          if !Cdglobals.plain_comments then Output.char '$'
          else escaped_html lexbuf; comment lexbuf )
# 4224 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1054 "tools/coqdoc/cpretty.mll"
      ( if !Cdglobals.parse_comments
        then
          (if !Cdglobals.plain_comments then Output.char '#'; Output.char '#');
        comment lexbuf )
# 4232 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1058 "tools/coqdoc/cpretty.mll"
         ( false )
# 4237 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1059 "tools/coqdoc/cpretty.mll"
           ( if !Cdglobals.parse_comments
             then Output.indentation (fst (count_spaces (lexeme lexbuf)));
             comment lexbuf )
# 4244 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1062 "tools/coqdoc/cpretty.mll"
         ( if !Cdglobals.parse_comments
           then Output.line_break (); comment lexbuf )
# 4250 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1064 "tools/coqdoc/cpretty.mll"
         ( if !Cdglobals.parse_comments then Output.char (lexeme_char lexbuf 0);
           comment lexbuf )
# 4256 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and skip_to_dot lexbuf =
    __ocaml_lex_skip_to_dot_rec lexbuf 2260
and __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1068 "tools/coqdoc/cpretty.mll"
                  ( true )
# 4268 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1069 "tools/coqdoc/cpretty.mll"
                     ( false )
# 4273 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1070 "tools/coqdoc/cpretty.mll"
         ( comment_level := 1; ignore (comment lexbuf); skip_to_dot lexbuf )
# 4278 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1071 "tools/coqdoc/cpretty.mll"
      ( skip_to_dot lexbuf )
# 4283 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_skip_to_dot_rec lexbuf __ocaml_lex_state

and body_bol lexbuf =
    __ocaml_lex_body_bol_rec lexbuf 2269
and __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1075 "tools/coqdoc/cpretty.mll"
      ( Output.indentation (fst (count_spaces (lexeme lexbuf))); body lexbuf )
# 4295 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1076 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; Output.indentation 0; body lexbuf )
# 4300 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_body_bol_rec lexbuf __ocaml_lex_state

and body lexbuf =
    __ocaml_lex_body_rec lexbuf 2272
and __ocaml_lex_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1079 "tools/coqdoc/cpretty.mll"
       (Tokens.flush_sublexer(); Output.line_break(); new_line lexbuf; body_bol lexbuf)
# 4312 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1081 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
            let s = lexeme lexbuf in
            let nlsp,s = remove_newline s in
            let _,isp = count_spaces s in
            let loc = lexeme_start lexbuf + nlsp + isp in
            Output.sublexer ']' loc;
            Output.sublexer ']' (loc+1);
            Tokens.flush_sublexer();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4333 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1099 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        if not !formatted then
          begin
	    let loc = lexeme_start lexbuf in
	    Output.sublexer ']' loc;
	    Output.sublexer ']' (loc+1);
	    Tokens.flush_sublexer();
	    Output.line_break();
            body lexbuf
          end
        else
          begin
            Output.paragraph ();
            true
          end )
# 4352 "tools/coqdoc/cpretty.ml"

  | 3 ->
# 1114 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); false )
# 4357 "tools/coqdoc/cpretty.ml"

  | 4 ->
# 1116 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.'; Output.line_break();
	  if not !formatted then true else body_bol lexbuf )
# 4363 "tools/coqdoc/cpretty.ml"

  | 5 ->
# 1119 "tools/coqdoc/cpretty.mll"
 ( Tokens.flush_sublexer(); Output.char '.';
        if not !formatted then
          begin
            eprintf "Error: stray ]] at %d\n"  (lexeme_start lexbuf);
            flush stderr;
            exit 1
          end
	  else
          begin
            Output.paragraph ();
            true
          end
      )
# 4380 "tools/coqdoc/cpretty.ml"

  | 6 ->
# 1133 "tools/coqdoc/cpretty.mll"
        ( Tokens.flush_sublexer(); Output.char '.'; Output.char ' ';
	  if not !formatted then false else body lexbuf )
# 4386 "tools/coqdoc/cpretty.ml"

  | 7 ->
# 1136 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.end_coq (); Output.start_doc ();
	let eol = doc_bol lexbuf in
	  Output.end_doc (); Output.start_coq ();
	  if eol then body_bol lexbuf else body lexbuf )
# 4394 "tools/coqdoc/cpretty.ml"

  | 8 ->
# 1140 "tools/coqdoc/cpretty.mll"
         ( Tokens.flush_sublexer(); comment_level := 1;
	   if !Cdglobals.parse_comments then Output.start_comment ();
	   let eol = comment lexbuf in
	     if eol
	     then begin if not !Cdglobals.parse_comments then Output.line_break(); body_bol lexbuf end
	     else body lexbuf )
# 4404 "tools/coqdoc/cpretty.ml"

  | 9 ->
# 1147 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) None;
	start_notation_string lexbuf )
# 4411 "tools/coqdoc/cpretty.ml"

  | 10 ->
# 1151 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer();
        Output.ident (lexeme lexbuf) (Some (lexeme_start lexbuf));
	body lexbuf )
# 4418 "tools/coqdoc/cpretty.ml"

  | 11 ->
# 1155 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '.'; Output.char '.';
        body lexbuf )
# 4424 "tools/coqdoc/cpretty.ml"

  | 12 ->
# 1158 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"';
        string lexbuf;
        body lexbuf )
# 4431 "tools/coqdoc/cpretty.ml"

  | 13 ->
# 1162 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
        body lexbuf )
# 4437 "tools/coqdoc/cpretty.ml"

  | 14 ->
# 1165 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
	Output.sublexer c (lexeme_start lexbuf);
        body lexbuf )
# 4444 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_body_rec lexbuf __ocaml_lex_state

and start_notation_string lexbuf =
    __ocaml_lex_start_notation_string_rec lexbuf 2348
and __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1170 "tools/coqdoc/cpretty.mll"
          ( Tokens.flush_sublexer(); Output.char (lexeme_char lexbuf 0);
	    start_notation_string lexbuf )
# 4457 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1173 "tools/coqdoc/cpretty.mll"
      ( Output.sublexer '"' (lexeme_start lexbuf);
        notation_string lexbuf;
	body lexbuf )
# 4464 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1177 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf; body lexbuf )
# 4469 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_start_notation_string_rec lexbuf __ocaml_lex_state

and notation_string lexbuf =
    __ocaml_lex_notation_string_rec lexbuf 2352
and __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1181 "tools/coqdoc/cpretty.mll"
      ( Output.char '"'; Output.char '"'; (* Unlikely! *)
        notation_string lexbuf )
# 4482 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1184 "tools/coqdoc/cpretty.mll"
      ( Tokens.flush_sublexer(); Output.char '"' )
# 4487 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1185 "tools/coqdoc/cpretty.mll"
      ( let c = lexeme_char lexbuf 0 in
        Output.sublexer c (lexeme_start lexbuf);
        notation_string lexbuf )
# 4494 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_notation_string_rec lexbuf __ocaml_lex_state

and string lexbuf =
    __ocaml_lex_string_rec lexbuf 2356
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1190 "tools/coqdoc/cpretty.mll"
           ( Output.char '"'; Output.char '"'; string lexbuf )
# 4506 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1191 "tools/coqdoc/cpretty.mll"
           ( Output.char '"' )
# 4511 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1192 "tools/coqdoc/cpretty.mll"
           ( let c = lexeme_char lexbuf 0 in Output.char c; string lexbuf )
# 4516 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and skip_hide lexbuf =
    __ocaml_lex_skip_hide_rec lexbuf 2360
and __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1195 "tools/coqdoc/cpretty.mll"
                   ( () )
# 4528 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1196 "tools/coqdoc/cpretty.mll"
      ( skip_hide lexbuf )
# 4533 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_skip_hide_rec lexbuf __ocaml_lex_state

and printing_token_body lexbuf =
    __ocaml_lex_printing_token_body_rec lexbuf 2376
and __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1202 "tools/coqdoc/cpretty.mll"
 ( let s = Buffer.contents token_buffer in
	  Buffer.clear token_buffer;
	  s )
# 4547 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1205 "tools/coqdoc/cpretty.mll"
        ( Buffer.add_string token_buffer (lexeme lexbuf);
	  printing_token_body lexbuf )
# 4553 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_printing_token_body_rec lexbuf __ocaml_lex_state

and inf_rules indents lexbuf =
    __ocaml_lex_inf_rules_rec indents lexbuf 2382
and __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1212 "tools/coqdoc/cpretty.mll"
      ( inf_rules indents lexbuf )
# 4565 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1214 "tools/coqdoc/cpretty.mll"
      ( match indents with
        | Some ls -> doc_list_bol ls lexbuf
        | None -> doc_bol lexbuf )
# 4572 "tools/coqdoc/cpretty.ml"

  | 2 ->
# 1218 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;  (* anything else must be the first line in a rule *)
        inf_rules_assumptions indents [] lexbuf)
# 4578 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_inf_rules_rec indents lexbuf __ocaml_lex_state

and inf_rules_assumptions indents assumptions lexbuf =
    __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf 2394
and __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1227 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let dashes_and_name = 
               cut_head_tail_spaces (String.sub line 0 (String.length line - 1))
        in
        let ldn = String.length dashes_and_name in
        let (dashes,name) = 
          try (let i = String.index dashes_and_name ' ' in
               let d = String.sub dashes_and_name 0 i in
               let n = cut_head_tail_spaces 
                        (String.sub dashes_and_name (i+1) (ldn-i-1))
               in
                 (d, Some n))
          with _ -> (dashes_and_name, None)
            
        in
          inf_rules_conclusion indents (List.rev assumptions) 
                               (spaces, dashes, name) [] lexbuf )
# 4607 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1246 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let assumption = cut_head_tail_spaces 
                            (String.sub line 0 (String.length line - 1))
        in
          inf_rules_assumptions indents ((spaces,assumption)::assumptions)
                                lexbuf )
# 4618 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_inf_rules_assumptions_rec indents assumptions lexbuf __ocaml_lex_state

and inf_rules_conclusion indents assumptions middle conclusions lexbuf =
    __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf 2404
and __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1260 "tools/coqdoc/cpretty.mll"
      ( backtrack lexbuf;
        Output.inf_rule assumptions middle (List.rev conclusions);
        inf_rules indents lexbuf )
# 4632 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1264 "tools/coqdoc/cpretty.mll"
      ( let line = lexeme lexbuf in
        let (spaces,_) = count_spaces line in
        let conc = cut_head_tail_spaces (String.sub line 0 
                                                    (String.length line - 1))
        in
          inf_rules_conclusion indents assumptions middle 
                               ((spaces,conc) :: conclusions) lexbuf
      )
# 4644 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_inf_rules_conclusion_rec indents assumptions middle conclusions lexbuf __ocaml_lex_state

and st_start m lexbuf =
    __ocaml_lex_st_start_rec m lexbuf 2416
and __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1276 "tools/coqdoc/cpretty.mll"
      ( st_modname m lexbuf )
# 4656 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1278 "tools/coqdoc/cpretty.mll"
      ( None )
# 4661 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_st_start_rec m lexbuf __ocaml_lex_state

and st_modname m lexbuf =
    __ocaml_lex_st_modname_rec m lexbuf 2424
and __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1282 "tools/coqdoc/cpretty.mll"
      ( if subtitle m (lexeme lexbuf) then
          st_subtitle lexbuf
        else
          None
      )
# 4677 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1288 "tools/coqdoc/cpretty.mll"
      ( None )
# 4682 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_st_modname_rec m lexbuf __ocaml_lex_state

and st_subtitle lexbuf =
    __ocaml_lex_st_subtitle_rec lexbuf 2468
and __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1292 "tools/coqdoc/cpretty.mll"
      ( let st = lexeme lexbuf in
        let i = try Str.search_forward (Str.regexp "\\**)") st 0 with
                   Not_found ->
                     (eprintf "unterminated comment at beginning of file\n";
                      exit 1)
        in
          Some (cut_head_tail_spaces (String.sub st 0 i))
      )
# 4701 "tools/coqdoc/cpretty.ml"

  | 1 ->
# 1301 "tools/coqdoc/cpretty.mll"
      ( None )
# 4706 "tools/coqdoc/cpretty.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_st_subtitle_rec lexbuf __ocaml_lex_state

;;

# 1304 "tools/coqdoc/cpretty.mll"
 
  let coq_file f m =
    reset ();
    let c = open_in f in
    let lb = from_channel c in
      (Index.current_library := m;
       Output.initialize ();
       Output.start_module ();
       Output.start_coq (); coq_bol lb; Output.end_coq ();
       close_in c)

  let detect_subtitle f m =
    let c = open_in f in
    let lb = from_channel c in
    let sub = st_start m lb in
      close_in c;
      sub

# 4732 "tools/coqdoc/cpretty.ml"
